# Copyright 2015 by Nedim Sabic (RabbitStack)
# All Rights Reserved.
# http://rabbitstack.github.io

# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from _ctypes import byref
from ctypes import c_ulong
import os

from fibratus.apidefs.sys import CONSOLE_SCREEN_BUFFER_INFO, INVALID_HANDLE_VALUE, get_std_handle, STD_OUTPUT_HANDLE, \
    create_console_screen_buffer, GENERIC_READ, GENERIC_WRITE, FILE_SHARE_READ, FILE_SHARE_WRITE, \
    CONSOLE_TEXTMODE_BUFFER, get_console_screen_buffer_info, COORD, SMALL_RECT, CHAR_INFO, \
    set_console_active_screen_buffer, write_console_output, CURSOR_INFO, get_console_cursor_info, \
    set_console_cursor_info, write_console_unicode
from fibratus.errors import TermInitializationError


HIGH_INTENSITY = 0x0008

# terminal colors
BLACK = 0x0000
DARK_BLUE = 0x0001
DARK_GREEN = 0x0002
DARK_RED = 0x0004
GRAY = DARK_BLUE | DARK_GREEN | DARK_RED
DARK_YELLOW = DARK_RED | DARK_GREEN
DARK_PURPLE = DARK_RED | DARK_BLUE
DARK_CYAN = DARK_GREEN | DARK_BLUE
LIGHT_WHITE = GRAY | HIGH_INTENSITY


class AnsiTerm(object):
    """Terminal's low level interface.

    Provides a set of methods to interact
    with the Windows terminals. By writing the chars
    directly to the screen buffer can prevent the
    annoying screen flickering.
    """

    def __init__(self):
        """Creates a new instance of the terminal.
        """
        self._cursor_info = CURSOR_INFO()
        self._console = INVALID_HANDLE_VALUE
        self._framebuffer = INVALID_HANDLE_VALUE
        self._char_buffer = None
        self._cols = 0
        self._rows = 0
        self._rect = None
        self._coord = COORD(0, 0)
        self._size = COORD(0, 0)

    def setup_console(self):
        """Initializes the screen frame buffer.

        Swaps the current screen buffer with a
        brand new created back buffer where the
        characters can be written to the flicker-free
        rectangular region.

        """
        self._console = get_std_handle(STD_OUTPUT_HANDLE)
        # could not get the standard
        # console handle, raise an exception
        if self._console == INVALID_HANDLE_VALUE:
            raise TermInitializationError()

        buffer_info = CONSOLE_SCREEN_BUFFER_INFO()
        get_console_screen_buffer_info(self._console, byref(buffer_info))
        get_console_cursor_info(self._console, byref(self._cursor_info))
        self._cursor_info.visible = False

        self._cols = buffer_info.size.x
        self._rows = buffer_info.size.y
        self._size = COORD(self._cols, self._rows)
        self._rect = SMALL_RECT(0, 0, self._cols - 1, self._rows - 1)
        self._char_buffer = (CHAR_INFO * (self._size.x * self._size.y))()
        self._framebuffer = create_console_screen_buffer(GENERIC_READ | GENERIC_WRITE,
                                                         FILE_SHARE_READ | FILE_SHARE_WRITE,
                                                         None,
                                                         CONSOLE_TEXTMODE_BUFFER,
                                                         None)
        if self._framebuffer == INVALID_HANDLE_VALUE:
            raise TermInitializationError()
        # hide the cursor and swap
        # the console active screen buffer
        set_console_cursor_info(self._framebuffer, byref(self._cursor_info))
        set_console_active_screen_buffer(self._framebuffer)

    def restore_console(self):
        if self._console:
            set_console_active_screen_buffer(self._console)
            self._cursor_info.visible = True
            set_console_cursor_info(self._console, byref(self._cursor_info))

    def write_output(self, charseq, color=LIGHT_WHITE):
        """Writes character and color attribute data to the frame buffer.

        The data to be written is taken from a correspondingly sized rectangular
        block at a specified location in the source buffer.

        Parameters
        ----------

        charseq: str
            the sequence of characters to be written on the frame buffer

        color: int
            the terminal output color
        """

        col = 0
        x = 0
        crlf = False

        if not charseq or len(charseq) <= 0:
            return

        try:
            for char in charseq:
                if char == '\n':
                    crlf = True
                col += 1
                # the last column has been reached.
                # If there was a carriage return
                # then stop the iteration
                if col == self._cols:
                    col = 0
                    if crlf:
                        crlf = False
                        continue

                if crlf:
                    crlf = False
                    space = col
                    # keep filling the rectangle with spaces
                    # until we reach the last column
                    while space <= self._cols:
                        self._char_buffer[space - 1].char.unicode_char = ' '
                        space += 1
                        x += 1
                    # reset the column and
                    # stop the current iteration
                    col = 0
                    continue
                self._char_buffer[x].char.unicode_char = char
                self._char_buffer[x].attributes = color
                x += 1
        except IndexError:
            pass
        # write the character attribute data
        # to the screen buffer
        write_console_output(self._framebuffer,
                             self._char_buffer,
                             self._size,
                             self._coord,
                             byref(self._rect))

    def write_console(self, charseq):
        """Writes a string to a console frame buffer
        beginning at the current cursor location.

        charseq: str
            the string to be written on the frame buffer
        """
        write_console_unicode(self._framebuffer, charseq, len(charseq), byref(c_ulong()), None)

    def cls(self):
        """Clears the current screen buffer.
        """
        for y in range(self._rows):
            for x in range(self._cols):
                i = (y * self._cols) + x
                self._char_buffer[i].char.unicode_char = ' '
        write_console_output(self._framebuffer,
                             self._char_buffer,
                             self._coord,
                             self._size,
                             byref(self._rect))




# Copyright 2016 by Nedim Sabic (RabbitStack)
# http://rabbitstack.github.io
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from .windows cimport ULONG
from .etw cimport *

cdef extern from "tdh.h":

    ctypedef ULONG TDHAPI

    ctypedef enum TDH_IN_TYPE:
        TDH_INTYPE_NULL = 0
        TDH_INTYPE_UNICODESTRING = 1
        TDH_INTYPE_ANSISTRING = 2
        TDH_INTYPE_INT8 = 3
        TDH_INTYPE_UINT8 = 4
        TDH_INTYPE_INT16 = 5
        TDH_INTYPE_UINT16 = 6
        TDH_INTYPE_INT32 = 7
        TDH_INTYPE_UINT32 = 8
        TDH_INTYPE_INT64 = 9
        TDH_INTYPE_UINT64 = 10
        TDH_INTYPE_FLOAT = 11
        TDH_INTYPE_DOUBLE = 12
        TDH_INTYPE_BOOLEAN = 13
        TDH_INTYPE_BINARY = 14
        TDH_INTYPE_GUID = 15
        TDH_INTYPE_POINTER = 16
        TDH_INTYPE_FILETIME = 17
        TDH_INTYPE_SYSTEMTIME = 18
        TDH_INTYPE_SID = 19
        TDH_INTYPE_HEXINT32 = 20
        TDH_INTYPE_HEXINT64 = 21
        TDH_INTYPE_COUNTEDSTRING = 300
        TDH_INTYPE_COUNTEDANSISTRING = 301
        TDH_INTYPE_REVERSEDCOUNTEDSTRING = 302
        TDH_INTYPE_REVERSEDCOUNTEDANSISTRING = 303
        TDH_INTYPE_NONNULLTERMINATEDSTRING = 304
        TDH_INTYPE_NONNULLTERMINATEDANSISTRING = 305
        TDH_INTYPE_UNICODECHAR = 306
        TDH_INTYPE_ANSICHAR = 307
        TDH_INTYPE_SIZET = 308
        TDH_INTYPE_HEXDUMP = 309
        TDH_INTYPE_WBEMSID = 310

    ctypedef enum TDH_OUT_TYPE:
        TDH_OUTTYPE_NULL = 0
        TDH_OUTTYPE_STRING = 1
        TDH_OUTTYPE_DATETIME = 2
        TDH_OUTTYPE_BYTE = 3
        TDH_OUTTYPE_UNSIGNEDBYTE = 4
        TDH_OUTTYPE_SHORT = 5
        TDH_OUTTYPE_UNSIGNEDSHORT  = 6
        TDH_OUTTYPE_INT = 6
        TDH_OUTTYPE_UNSIGNEDINT = 7
        TDH_OUTTYPE_LONG  = 8
        TDH_OUTTYPE_UNSIGNEDLONG = 9
        TDH_OUTTYPE_FLOAT = 10
        TDH_OUTTYPE_DOUBLE = 11
        TDH_OUTTYPE_BOOLEAN = 12
        TDH_OUTTYPE_GUID = 13
        TDH_OUTTYPE_HEXBINARY = 14
        TDH_OUTTYPE_HEXINT8 = 15
        TDH_OUTTYPE_HEXINT16 = 16
        TDH_OUTTYPE_HEXINT32 = 17
        TDH_OUTTYPE_HEXINT64 = 18
        TDH_OUTTYPE_PID  = 19
        TDH_OUTTYPE_TID  = 20
        TDH_OUTTYPE_PORT  = 21
        TDH_OUTTYPE_IPV4  = 22
        TDH_OUTTYPE_IPV6  = 23
        TDH_OUTTYPE_SOCKETADDRESS  = 24
        TDH_OUTTYPE_CIMDATETIME  = 25
        TDH_OUTTYPE_ETWTIME  = 26
        TDH_OUTTYPE_XML  = 27
        TDH_OUTYTPE_ERRORCODE  = 28,
        TDH_OUTTYPE_REDUCEDSTRING = 300

    ctypedef enum PROPERTY_FLAGS:
        PropertyStruct        = 0x1
        PropertyParamLength   = 0x2
        PropertyParamCount    = 0x4
        PropertyWBEMXmlFragment = 0x8
        PropertyParamFixedLength = 0x10


    ctypedef struct NON_STRUCT_TYPE:
        USHORT in_type "InType"
        USHORT out_type "OutType"

    ctypedef struct EVENT_PROPERTY_INFO:
        ULONG name_offset "NameOffset"
        NON_STRUCT_TYPE non_struct_type "nonStructType"

    ctypedef struct TRACE_PROVIDER_INFO:
        GUID ProviderGuid
        USHORT PropertyCount


    ctypedef struct TDH_CONTEXT:
        pass

    ctypedef struct PROPERTY_DATA_DESCRIPTOR:
        ULONGLONG property_name "PropertyName"
        ULONG     array_index "ArrayIndex"
        ULONG     reserved "Reserved"


    ctypedef struct TRACE_EVENT_INFO:
        GUID ProviderGuid
        GUID event_guid "EventGuid"
        ULONG ProviderNameOffset
        ULONG OpcodeNameOffset
        ULONG PropertyCount
        ULONG property_count "TopLevelPropertyCount"
        EVENT_PROPERTY_INFO properties "EventPropertyInfoArray"[1]


    TDHAPI tdh_get_event_information "TdhGetEventInformation"(EVENT_RECORD* e, ULONG cc,
                                                              TDH_CONTEXT* ctx,
                                                              TRACE_EVENT_INFO* buf,
                                                              ULONG* buf_size) nogil

    ULONG  tdh_get_property_size "TdhGetPropertySize"(EVENT_RECORD* e, ULONG cc,
                                                      TDH_CONTEXT* ctx,
                                                      ULONG count,
                                                      PROPERTY_DATA_DESCRIPTOR* descriptor,
                                                      ULONG *size) nogil

    ULONG  tdh_get_property "TdhGetProperty"(EVENT_RECORD* e, ULONG cc,
                                             TDH_CONTEXT* ctx,
                                             ULONG count,
                                             PROPERTY_DATA_DESCRIPTOR* descriptor,
                                             ULONG buf_size,
                                             BYTE* buf) nogil



# Copyright 2016 by Nedim Sabic (RabbitStack)
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import elasticsearch
import elasticsearch.helpers

from fibratus.errors import InvalidPayloadError
from fibratus.output.base import Output


class ElasticsearchOutput(Output):

    def __init__(self, **kwargs):
        """Creates an instance of the Elasticsearch output adapter.

        Parameters
        ----------

        kwargs: dict
            Elasticsearch cluster configuration
        """
        Output.__init__(self)

        hosts = kwargs.pop('hosts', [])
        self._hosts = [dict(host=host.split(':')[0], port=int(host.split(':')[1])) for host in hosts]
        self._index_name = kwargs.pop('index', None)
        self._document_type = kwargs.pop('document', None)
        self._bulk = kwargs.pop('bulk', False)
        self._username = kwargs.pop('username', None)
        self._password = kwargs.pop('password', None)
        self._config = {}
        if self._username and self._password:
            self._config['http_auth'] = (self._username, self._password,)
        self._config['use_ssl'] = kwargs.pop('ssl', False)
        self._elasticsearch = None

    def emit(self, body, **kwargs):
        if not self._elasticsearch:
            self._elasticsearch = elasticsearch.Elasticsearch(self._hosts, **self._config)
        if self._bulk:
            if not isinstance(body, list):
                raise InvalidPayloadError('invalid payload for bulk indexing. '
                                          'list expected but %s found'
                                          % type(body))
        else:
            if not isinstance(body, dict):
                raise InvalidPayloadError('invalid payload for document. '
                                          'dict expected but %s found'
                                          % type(body))

        self._index_name = kwargs.pop('index', self._index_name)
        if self._bulk:
            actions = [dict(_index=self._index_name, _type=self._document_type, _source=b) for b in body]
            elasticsearch.helpers.bulk(self._elasticsearch, actions)
        else:
            self._elasticsearch.index(self._index_name, self._document_type, body=body)

    @property
    def hosts(self):
        return self._hosts

    @property
    def index_name(self):
        return self._index_name

    @property
    def document_type(self):
        return self._document_type

    @property
    def bulk(self):
        return self._bulk



# Copyright 2016 by Nedim Sabic (RabbitStack)
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from unittest.mock import patch

import pytest

from fibratus.apidefs.sys import STD_OUTPUT_HANDLE, GENERIC_READ, GENERIC_WRITE, \
    FILE_SHARE_READ, CONSOLE_TEXTMODE_BUFFER, FILE_SHARE_WRITE, INVALID_HANDLE_VALUE, COORD, SMALL_RECT, CHAR_INFO, \
    CONSOLE_SCREEN_BUFFER_INFO
from fibratus.errors import TermInitializationError
from fibratus.term import AnsiTerm


class TestAnsiTerm(object):

    @patch('fibratus.term.get_std_handle', return_value=1)
    @patch('fibratus.term.get_console_screen_buffer_info')
    @patch('fibratus.term.get_console_cursor_info')
    @patch('fibratus.term.create_console_screen_buffer', return_value=2)
    @patch('fibratus.term.set_console_cursor_info')
    @patch('fibratus.term.set_console_active_screen_buffer')
    def test_setup_console(self, set_console_active_screen_buffer_mock,
                           set_console_cursor_info_mock,
                           create_console_screen_buffer_mock,
                           get_console_cursor_info_mock,
                           get_console_screen_buffer_info_mock,
                           get_std_handle_mock):

        with patch('fibratus.term.byref', side_effect=[1, 2, 3]):
            ansi_term = AnsiTerm()
            ansi_term.setup_console()

            get_std_handle_mock.assert_called_with(STD_OUTPUT_HANDLE)
            get_console_screen_buffer_info_mock.assert_called_with(1, 1)
            get_console_cursor_info_mock.assert_called_with(1, 2)

            create_console_screen_buffer_mock.assert_called_with(GENERIC_READ | GENERIC_WRITE,
                                                                 FILE_SHARE_READ | FILE_SHARE_WRITE,
                                                                 None,
                                                                 CONSOLE_TEXTMODE_BUFFER,
                                                                 None)

            set_console_cursor_info_mock.assert_called_with(2, 3)
            set_console_active_screen_buffer_mock.assert_called_with(2)

    @patch('fibratus.term.get_std_handle', return_value=INVALID_HANDLE_VALUE)
    def test_setup_console_invalid_std_console(self, get_std_handle_mock):

            ansi_term = AnsiTerm()
            with pytest.raises(TermInitializationError):
                ansi_term.setup_console()
                get_std_handle_mock.assert_called_with(STD_OUTPUT_HANDLE)

    @patch('fibratus.term.get_std_handle', return_value=1)
    @patch('fibratus.term.get_console_screen_buffer_info')
    @patch('fibratus.term.get_console_cursor_info')
    @patch('fibratus.term.create_console_screen_buffer', return_value=INVALID_HANDLE_VALUE)
    def test_setup_console_invalid_frame_buffer(self, create_console_screen_buffer_mock,
                                                get_console_cursor_info_mock,
                                                get_console_screen_buffer_info_mock,
                                                get_std_handle_mock):

            ansi_term = AnsiTerm()
            with pytest.raises(TermInitializationError):
                ansi_term.setup_console()
                create_console_screen_buffer_mock.assert_called_with(GENERIC_READ | GENERIC_WRITE,
                                                                     FILE_SHARE_READ | FILE_SHARE_WRITE,
                                                                     None,
                                                                     CONSOLE_TEXTMODE_BUFFER,
                                                                     None)

    @patch('fibratus.term.get_std_handle', return_value=1)
    @patch('fibratus.term.get_console_screen_buffer_info')
    @patch('fibratus.term.get_console_cursor_info')
    @patch('fibratus.term.create_console_screen_buffer', return_value=2)
    @patch('fibratus.term.set_console_cursor_info')
    @patch('fibratus.term.set_console_active_screen_buffer')
    def test_restore_console(self, set_console_active_screen_buffer_mock,
                             set_console_cursor_info_mock,
                             create_console_screen_buffer_mock,
                             get_console_cursor_info_mock,
                             get_console_screen_buffer_info_mock,
                             get_std_handle_mock):

        ansi_term = AnsiTerm()
        ansi_term.setup_console()
        with patch('fibratus.term.byref', return_value=2):
            ansi_term.restore_console()
            set_console_active_screen_buffer_mock.assert_called_with(1)
            set_console_cursor_info_mock.assert_called_with(1, 2)

    @patch('fibratus.term.get_std_handle', return_value=1)
    @patch('fibratus.term.get_console_screen_buffer_info')
    @patch('fibratus.term.get_console_cursor_info')
    @patch('fibratus.term.create_console_screen_buffer', return_value=2)
    @patch('fibratus.term.set_console_cursor_info')
    @patch('fibratus.term.set_console_active_screen_buffer')
    @patch('fibratus.term.write_console_output')
    def test_write_output(self, write_console_output_mock,
                          set_console_active_screen_buffer_mock,
                          set_console_cursor_info_mock,
                          create_console_screen_buffer_mock,
                          get_console_cursor_info_mock,
                          get_console_screen_buffer_info_mock,
                          get_std_handle_mock):
        ansi_term = AnsiTerm()

        buffer_info = CONSOLE_SCREEN_BUFFER_INFO()
        buffer_info.size.x = 200
        buffer_info.size.y = 300

        with patch.object(CONSOLE_SCREEN_BUFFER_INFO, '__new__', return_value=buffer_info):
            ansi_term.setup_console()
            ansi_term.write_output('Top inbound packets\n')
            write_console_output_mock.assert_called_once()





# Copyright 2015/2016 by Nedim Sabic (RabbitStack)
# http://rabbitstack.github.io
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import os
from ctypes import cast, byref

from fibratus.common import NA
from fibratus.handle import HandleType
from fibratus.kevent_types import *
from fibratus.apidefs.registry import *
from fibratus.apidefs.sys import malloc, free


class HiveParser(object):

    def __init__(self, kevent, thread_registry):
        self._kcblocks = {}
        self._kevent = kevent
        self.thread_registry = thread_registry
        self.hive_regexs = [
            r'(?i)(REGISTRY\\MACHINE\\SOFTWARE)(.*)',
            r'(?i)(REGISTRY\\MACHINE\\HARDWARE)(.*)',
            r'(?i)(REGISTRY\\MACHINE\\SECURITY)(.*)',
            r'(?i)(REGISTRY\\MACHINE\\SYSTEM)(.*)',
            r'(?i)(REGISTRY\\MACHINE\\SAM)(.*)',
            r'(?i)(REGISTRY\\USER\\.DEFAULT)(.*)',
            r'(?i)(REGISTRY\\USER\\S-.+?)\\(.*)',
            r'(?i)(REGISTRY\\USER)(.*)']
        self._reg_value_types = [v.value for v in ValueType]

    @property
    def kcblocks(self):
        return self._kcblocks

    def remove_kcb(self, key_handle):
        if key_handle in self._kcblocks:
            self._kcblocks.pop(key_handle)

    def add_kcb(self, kkcb):
        """Adds a key control block (KCB).

        Parameters
        ----------

        kkcb: dict
            metadata for the KCB

        """
        handle = kkcb.key_handle
        # index the KCB by key handle
        # we also save the process and
        # thread id which created the KCB
        kcb = Kcb(handle, kkcb.key_name,
                  kkcb.index,
                  kkcb.status,
                  kkcb.thread_id,
                  kkcb.process_id)
        self._kcblocks[handle] = kcb

    def parse_hive(self, ketype, regkevt):
        """Parses a hive from the registry kernel event.

        Hive is a logical group of keys, subkeys and values
        which are commonly called as nodes.

        Parameters
        ----------

        ketype: tuple
            kernel event type
        regkevt: dict
            kernel registry event payload as forwarded from the
            event stream collector
        """
        hive = NA
        key = regkevt.key_name
        status = regkevt.status
        tid = regkevt.thread_id
        pid = regkevt.process_id
        index = regkevt.index

        self._kevent.tid = tid
        self._kevent.pid = pid

        # if the node handle (KCB handle) is equal to 0
        # we have the full node name. Otherwise
        # we have to query the key control blocks
        # to found the full node name
        handle = regkevt.key_handle
        if handle == 0:
            # find the hive by applying
            # the regular expression
            hive, key = self._dissect_hive(key)
        else:
            if handle in self._kcblocks:
                # KCB found. Concatenate the
                # full node path
                kcb = self._kcblocks[handle]
                full_path = '%s\%s' % (kcb.key, key)
                hive, key = self._dissect_hive(full_path)
            else:
                # we missed the KCB creation
                # lookup the handles
                # to find the key name
                thread = self.thread_registry.get_thread(pid)
                if thread:
                    key_handles = [kh for kh in thread.handles if kh.handle_type is not None and
                                   kh.handle_type == HandleType.KEY]
                    for khandle in key_handles:
                        if ketype in [REG_CREATE_KEY,
                                      REG_DELETE_KEY,
                                      REG_OPEN_KEY,
                                      REG_QUERY_KEY]:
                            # try to find the match of the key name
                            # from registry key handle name.
                            # Replace the backslash to prevent
                            # bogus escape exceptions
                            khandle_name = khandle.name
                            f = re.findall(r"%s" % key.replace('\\', '_'),
                                           khandle_name.replace('\\', '_'))
                            if len(f) > 0:
                                hive, key = self._dissect_hive(khandle_name)
                                kcb = Kcb(handle,
                                          khandle_name,
                                          index,
                                          status,
                                          tid,
                                          pid)
                                self._kcblocks[handle] = kcb
                                break

        if hive == NA:
            # set the unknown hive and
            # the partial node name
            key = '..\%s' % key

        if ketype in [REG_CREATE_KEY,
                      REG_DELETE_KEY,
                      REG_OPEN_KEY,
                      REG_QUERY_KEY]:
            params = {
                'hive': hive,
                'key': key,
                'status': status,
                'tid': tid,
                'pid': pid
            }
            self._kevent.params = params
        elif ketype in [REG_SET_VALUE,
                        REG_DELETE_VALUE,
                        REG_QUERY_VALUE]:
            if ketype == REG_SET_VALUE or ketype == REG_QUERY_VALUE:
                # we have the hive and the subkey
                # including the registry value name
                # which means we are able to query the content
                # of the registry value
                if hive != NA and not key.startswith('..'):
                    # resolve the root key name
                    # from the registry hive
                    hkey = self._hive_to_hkey(hive)
                    subkey, value_name = os.path.split(key)
                    # get the value data and value type
                    # from the registry
                    value, value_type = self._query_value(hkey,
                                                          subkey,
                                                          value_name)
                    self._kevent.params = dict(hive=hive, key=key,
                                               value_type=value_type,
                                               value=value,
                                               status=status,
                                               tid=tid,
                                               pid=pid)
                else:
                    self._kevent.params = dict(hive=hive, key=key,
                                               value_type=NA,
                                               value=NA, status=status,
                                               tid=tid, pid=pid)

            else:
                self._kevent.params = dict(hive=hive, key=key,
                                           status=status,
                                           tid=tid,
                                           pid=pid)

    def _query_value(self, hkey, subkey, value_name):
        """Get value content and value type from registry.

        Parameters
        ----------

        hkey: HKEY
            handle to registry root key
        subkey: str
            path representing the subkey
        value:
            the name of the value
        """
        if not hkey:
            return NA, NA
        value_type = c_ulong()
        buff = malloc(MAX_BUFFER_SIZE)
        buff_size = c_ulong(MAX_BUFFER_SIZE)

        status = reg_get_value(hkey, c_wchar_p(subkey),
                               c_wchar_p(value_name),
                               RRF_RT_ANY,
                               byref(value_type),
                               buff, byref(buff_size))
        if status == ERROR_SUCCESS:
            value = cast(buff, c_wchar_p).value
            value_type = value_type.value
            if value_type in self._reg_value_types:
                if value_type == ValueType.REG_BINARY.value:
                    value = '<binary>'
                [value_type] = [v.name for v in ValueType if v.value == value_type]
            else:
                value_type = ValueType.REG_NONE.name
            free(buff)
            return value, value_type
        else:
            free(buff)
            return NA, NA

    def _dissect_hive(self, key_name):
        """Extracts the hive name and the subkey from the key path.

        Parameters
        ----------

        key_name: str
            key path from whom the hive
            can be resolved
        """
        for rx in self.hive_regexs:
            # for each regex match it
            # against key path
            m = re.search(rx, key_name)
            if m and len(m.groups()) > 0:
                hive = m.group(1).upper()
                # hive found, now try
                # to get the node path
                if len(m.groups()) > 1:
                    node = m.group(2)
                    if node:
                        # because the hive contains the
                        # child node of the registry subkey
                        # we have to include it
                        _, hive_child = os.path.split(hive)
                        if not node.startswith('\\'):
                            node = '\\%s' % node
                        node = '%s%s' % (hive_child, node)
                        return hive.replace('\\', '_'), node
                return hive.replace('\\', '_'), key_name
        return key_name, key_name

    def _hive_to_hkey(self, hive):
        if re.match(r'(?i).*MACHINE.*', hive):
            return HKEY_LOCAL_MACHINE
        elif re.match(r'(?i).*USER_S-.*|\.DEFAULT', hive):
            return HKEY_USERS
        else:
            return None


class Kcb(object):
    """The container for the Key Control Block data.
    """
    def __init__(self, handle, key, index,
                 status, tid, pid):
        self._handle = handle
        self._key = key
        self._index = index
        self._status = status
        self._thread_id = tid
        self._process_id = pid

    @property
    def key(self):
        return self._key






# Copyright 2015 by Nedim Sabic (RabbitStack)
# All Rights Reserved.
# http://rabbitstack.github.io

# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from ctypes.wintypes import HKEY, DWORD, LPDWORD, LONG, LPCWSTR
from enum import Enum

from fibratus.apidefs.cdefs import *
import fibratus.apidefs.declarer as declarer


# query type flags
RRF_RT_ANY = 0x0000ffff

# reserved key handles
HKEY_CLASSES_ROOT = HKEY(0x80000000)
HKEY_CURRENT_USER = HKEY(0x80000001)
HKEY_LOCAL_MACHINE = HKEY(0x80000002)
HKEY_USERS = HKEY(0x80000003)

MAX_BUFFER_SIZE = 4096
reg_get_value = declarer.declare(declarer.ADVAPI, 'RegGetValueW',
                                 [HKEY, LPCWSTR, LPCWSTR,
                                 DWORD, LPDWORD, PVOID, LPDWORD],
                                 LONG)


class ValueType(Enum):
    REG_NONE = 0
    REG_SZ = 1
    REG_EXPAND_SZ = 2
    REG_BINARY = 3
    REG_DWORD = 4
    REG_DWORD_BIG_ENDIAN = 5
    REG_LINK = 6
    REG_MULTI_SZ = 7
    REG_RESOURCE_LIST = 8
    REG_FULL_RESOURCE_DESCRIPTOR = 9
    REG_RESOURCE_REQUIREMENTS_LIST = 10
    REG_QWORD = 11



# Copyright 2016 by Nedim Sabic (RabbitStack)
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os
import sys
import anyconfig
from fibratus.common import panic, DotD as ddict
from pykwalify.core import Core


class YamlConfig(object):
    """YAML based configuration reader.

    Reads the configuration from YAML file, and ensures the content satisfies the structure
    as defined in the schema file.
    """

    def __init__(self, config_path=None):
        self._default_config_path = os.path.join(os.path.expanduser('~'), '.fibratus', 'fibratus.yml')
        self._default_schema_path = os.path.join(os.path.expanduser('~'), '.fibratus', 'schema.yml')
        self.path = config_path or os.getenv('FIBRATUS_CONFIG_PATH', self._default_config_path)
        self._yaml = None

    def load(self, validate=True):
        schema_file = os.path.join(sys._MEIPASS, 'schema.yml') \
            if hasattr(sys, '_MEIPASS') else self._default_schema_path
        try:
            self._yaml = anyconfig.load(self.path, ignore_missing=False)
        except FileNotFoundError:
            panic('ERROR - %s configuration file does not exist' % self.path)
        if validate:
            validator = Core(source_file=self.path, schema_files=[schema_file])
            validator.validate(raise_exception=True)

    @property
    def image_meta(self):
        return ddict(self._yaml.pop('image_meta', {}))

    @property
    def skips(self):
        return ddict(self._yaml.pop('skips', {}))

    @property
    def outputs(self):
        return self._yaml.pop('output', None)

    @property
    def bindings(self):
        return self._yaml.pop('binding', None)

    @property
    def yaml(self):
        return self._yaml

    @property
    def default_config_path(self):
        return self._default_config_path

    @default_config_path.setter
    def default_config_path(self, path):
        self._default_config_path = path

    @property
    def default_schema_path(self):
        return self._default_schema_path

    @default_schema_path.setter
    def default_schema_path(self, path):
        self._default_schema_path = path

    @property
    def config_path(self):
        return self.path


# Copyright 2015 by Nedim Sabic (RabbitStack)
# All Rights Reserved.
# http://rabbitstack.github.io

# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from _ctypes import POINTER, byref, addressof
from ctypes import cast, c_ulong, c_wchar_p
from ctypes.wintypes import HANDLE, ULONG
from enum import Enum

from fibratus.common import DotD as ddict
from fibratus.apidefs.cdefs import STATUS_INFO_LENGTH_MISMATCH, STATUS_SUCCESS, ERROR_SUCCESS, \
    UNICODE_STRING
from fibratus.apidefs.process import open_process, PROCESS_DUP_HANDLE, get_current_process
from fibratus.apidefs.registry import MAX_BUFFER_SIZE
from fibratus.apidefs.sys import zw_query_system_information, SYSTEM_HANDLE_INFORMATION_CLASS, \
    SYSTEM_HANDLE_INFORMATION, free, realloc, SYSTEM_HANDLE, malloc, duplicate_handle, nt_query_object, \
    PUBLIC_OBJECT_TYPE_INFORMATION, OBJECT_TYPE_INFORMATION, PUBLIC_OBJECT_NAME_INFORMATION, close_handle
from fibratus.errors import HandleEnumError


class HandleType(Enum):
    FILE = 0
    DIRECTORY = 1
    KEY = 2
    ALPC_PORT = 3
    SECTION = 4
    MUTANT = 5
    EVENT = 6
    DESKTOP = 7
    SEMAPHORE = 8
    TIMER = 9
    TOKEN = 10
    JOB = 11


class HandleRepository(object):
    """Stores open handle objects.
    """

    def __init__(self):
        self._object_buff_size = 0x1000
        self._object_types = {}
        # the object handles with these
        # masks shouldn't be queried,
        # otherwise the call could hang
        # the main thread
        self._nasty_access_masks = [0x120189,
                                    0x0012019f,
                                    0x1A019F]

        self._handle_types = [name for name, _ in HandleType.__members__.items()]
        self._buffers = []

    def query_handles(self, pid=None):
        raw_handles = self._enum_handles(pid)
        current_ps = HANDLE(get_current_process())
        handles = []
        # find the object handles for the process
        for _, handle in raw_handles.items():
            ps_handle = open_process(PROCESS_DUP_HANDLE,
                                     False,
                                     handle.pid)
            if ps_handle:
                handle_copy = HANDLE()
                # to query the object handle
                # we need to duplicate it in
                # the address space of the current process
                status = duplicate_handle(ps_handle,
                                          handle.handle,
                                          current_ps,
                                          byref(handle_copy),
                                          0, 0, 0)
                if status != ERROR_SUCCESS:
                    # get the object type
                    handle_type = self._query_handle(handle_copy,
                                                     PUBLIC_OBJECT_TYPE_INFORMATION,
                                                     OBJECT_TYPE_INFORMATION)
                    if handle_type:
                        handle_type = cast(handle_type.contents.type_name.buffer, c_wchar_p) \
                            .value \
                            .upper().replace(' ', '_')
                        # query for object name
                        # (file names, registry keys,
                        # sections, ALPC ports, etc)
                        # check the access mask to make
                        # sure `NtQueryObject` won't hang
                        if handle_type in self._handle_types and \
                                handle.access_mask not in self._nasty_access_masks:
                            handle_name = self._query_handle(handle_copy,
                                                             PUBLIC_OBJECT_NAME_INFORMATION,
                                                             UNICODE_STRING)
                            if handle_name:
                                handle_name = cast(handle_name.contents.buffer, c_wchar_p).value
                            handle_info = HandleInfo(handle.handle,
                                                     handle.obj,
                                                     HandleType(HandleType.__getattr__(handle_type)),
                                                     handle_name,
                                                     handle.pid)
                            handles.append(handle_info)

                    close_handle(handle_copy)
                close_handle(ps_handle)
        return handles

    def free_buffers(self):
        for buff in self._buffers:
            free(buff)

    def _enum_handles(self, process_id=None):
        """Enumerates handle information.

        Enumerates handle info on
        the start of the kernel capture.

        Returns a dictionary of handle's
        information including the handle id,
        access mask, and the process which owns
        the handle.
        """
        buff_size = MAX_BUFFER_SIZE
        size = c_ulong()
        # allocate the initial buffer
        buff = malloc(buff_size)
        handles = {}

        while True:
            status = zw_query_system_information(SYSTEM_HANDLE_INFORMATION_CLASS,
                                                 buff,
                                                 buff_size,
                                                 byref(size))
            if status == STATUS_INFO_LENGTH_MISMATCH:
                # the buffer is too small
                # increment the buffer size and try again
                buff_size += MAX_BUFFER_SIZE
            elif status == STATUS_SUCCESS:
                # cast the buffer to `SYSTEM_HANDLE_INFORMATION` struct
                # which contains an array of `SYSTEM_HANDLE` structures
                sys_handle_info = cast(buff, POINTER(SYSTEM_HANDLE_INFORMATION))
                sys_handle_info = sys_handle_info.contents
                handle_count = sys_handle_info.number_of_handles

                # resize the array size to the
                # actual number of file handles
                sys_handles = (SYSTEM_HANDLE * buff_size).from_address(addressof(sys_handle_info.handles))

                for i in range(handle_count):
                    sys_handle = sys_handles[i]
                    pid = sys_handle.process_id
                    handle = sys_handle.handle
                    obj = sys_handle.object
                    obj_type_index = sys_handle.object_type_number
                    access_mask = sys_handle.access_mask
                    if process_id and process_id == pid:
                        handles[obj] = ddict(pid=process_id,
                                             handle=handle,
                                             obj=obj,
                                             access_mask=access_mask,
                                             obj_type_index=obj_type_index)
                    elif process_id is None:
                        handles[obj] = ddict(pid=pid,
                                             handle=handle,
                                             obj=obj,
                                             access_mask=access_mask,
                                             obj_type_index=obj_type_index)
                break
            else:
                raise HandleEnumError(status)
            # reallocate the buffer
            buff = realloc(buff, buff_size)
        # free the buffer memory
        free(buff)

        return handles

    def _async_query_object(self):
        pass

    def _query_handle(self, handle, klass, object_info_type):
        """Gets the object handle info.

        Parameters
        ----------


        handle: HANDLE
            handle object
        klass: int
            the class of information to query
        object_info_type: Structure
            structure type which holds the handle info
        """
        buff = malloc(self._object_buff_size)
        rlen = ULONG()
        status = nt_query_object(handle,
                                 klass,
                                 buff,
                                 self._object_buff_size,
                                 byref(rlen))
        if status >= 0:
            info = cast(buff, POINTER(object_info_type))
            self._buffers.append(buff)
            return info
        else:
            # reallocate the buffer size
            # and try again
            buff = realloc(buff, rlen.value)
            status = nt_query_object(handle,
                                     klass,
                                     buff,
                                     self._object_buff_size,
                                     None)
            if status >= 0:
                info = cast(buff, POINTER(object_info_type))
                self._buffers.append(buff)
                return info
            else:
                free(buff)
                return None


class HandleInfo():
    """Saves the handle meta data.
    """

    def __init__(self, handle, obj, handle_type, name, pid):
        self._handle = handle
        self._obj = obj
        self._handle_type = handle_type
        self._name = name
        self._pid = pid

    @property
    def name(self):
        return self._name

    @property
    def handle_type(self):
        return self._handle_type

    @property
    def obj(self):
        return self._obj

    @property
    def pid(self):
        return self._pid

    @property
    def handle(self):
        return self._handle

    def __str__(self):
        return '%s type: [%s] object address: [%s] handle id: [%s] pid: [%s]' % \
               (self._name, self._handle_type,
                hex(self._obj),
                self._handle,
                self._pid)



# Copyright 2015 by Nedim Sabic (RabbitStack)
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import ctypes
import re


class GUID(ctypes.Structure):
    _DECOMPOSE_RE = re.compile('{([0-9A-F]{8})-([0-9A-F]{4})-([0-9A-F]{4})-([0-9A-F]{2})([0-9A-F]{2})-'
                               '([0-9A-F]{2})([0-9A-F]{2})([0-9A-F]{2})'
                               '([0-9A-F]{2})([0-9A-F]{2})([0-9A-F]{2})}', re.I)

    def __init__(self, guid_as_str=None):
        if guid_as_str:
            m = self._DECOMPOSE_RE.match(guid_as_str)
            g = [int(i, 16) for i in m.groups()]
            self.Data1 = g[0]
            self.Data2 = g[1]
            self.Data3 = g[2]
            for i in range(8):
                self.Data4[i] = g[3 + i]

    _fields_ = [("Data1", ctypes.c_ulong),
                ("Data2", ctypes.c_ushort),
                ("Data3", ctypes.c_ushort),
                ("Data4", ctypes.c_ubyte * 8)]

    def __str__(self):
        return "{%08x-%04x-%04x-%02x%02x-%02x%02x%02x%02x%02x%02x}" % \
               (self.Data1, self.Data2, self.Data3,
                self.Data4[0], self.Data4[1],
                self.Data4[2], self.Data4[3], self.Data4[4],
                self.Data4[5], self.Data4[6], self.Data4[7])



# Copyright 2015 by Nedim Sabic (RabbitStack)
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import inspect
import traceback
import os
import sys
from importlib.machinery import SourceFileLoader

from apscheduler.executors.pool import ThreadPoolExecutor
from apscheduler.schedulers.background import BackgroundScheduler
from apscheduler.triggers.interval import IntervalTrigger

from fibratus.common import DotD as ddict, Tabular
from fibratus.common import panic
from fibratus.errors import FilamentError, TermInitializationError
from fibratus.term import AnsiTerm


FILAMENTS_DIR = os.getenv('FILAMENTS_PATH', os.path.join(os.path.expanduser('~'), '.fibratus', 'filaments'))


class OutputAccessor(object):
    """An accessor for the output meta variable.

    It represents an output accessor which is injected into
    every filament module.
    """
    def __init__(self, output):
        self._output = output

    def emit(self, body, **kwargs):
        self._output.emit(body, **kwargs)


class Filament(object):
    """Filament initialization and execution engine.

    Filaments are lightweight Python modules which run
    on top of Fibratus. They are often used to enrich/extend the
    functionality of Fibratus by performing any type of logic
    (aggregations, groupings, filters, counters, etc) on the
    kernel event stream.

    """
    def __init__(self):
        """Builds a new instance of the filament.

        Attributes:
        ----------

        filament_module: module
            module which contains the filament logic
        """
        self._filament_module = None
        self._name = None
        self._filters = []
        self._cols = []
        self._tabular = None
        self._limit = 10
        self._interval = 1
        self._sort_by = None
        self._sort_desc = True
        self._logger = None
        self._ansi_term = AnsiTerm()
        self.scheduler = BackgroundScheduler()

    def load_filament(self, name):
        """Loads the filament module.

        Finds and loads the python module which
        holds the filament logic. It also looks up for
        some essential filament methods and raises an error
        if they can't be found.

        Parameters
        ----------
        name: str
            name of the filament to load

        """
        self._name = name
        Filament._assert_root_dir()
        filament_path = self._find_filament_path(name)
        if filament_path:
            loader = SourceFileLoader(name, filament_path)
            self._filament_module = loader.load_module()
            sys.path.append(FILAMENTS_DIR)
            doc = inspect.getdoc(self._filament_module)
            if not doc:
                raise FilamentError('Please provide a short '
                                    'description for the filament')

            on_next_kevent = self._find_filament_func('on_next_kevent')
            if on_next_kevent:
                if self._num_args(on_next_kevent) != 1:
                    raise FilamentError('Missing one argument on_next_kevent '
                                        'method on filament')
                self._initialize_funcs()
            else:
                raise FilamentError('Missing required on_next_kevent '
                                    'method on filament')
        else:
            raise FilamentError('%s filament not found' % name)

    def _initialize_funcs(self):
        """Setup the filament modules functions.

        Functions
        ---------

        set_filter: func
            accepts the comma separated list of kernel events
            for whose the filter should be applied
        set_interval: func
            establishes the fixed repeating interval in seconds
        columns: func
            configure the column set for the table
        add_row: func
            adds a new row to the table
        sort_by: func
            sorts the table by specific column
        """

        def set_filter(*args):
            self._filters = args
        self._filament_module.set_filter = set_filter

        def set_interval(interval):
            if not type(interval) is int:
                raise FilamentError('Interval must be an integer value')
            self._interval = interval
        self._filament_module.set_interval = set_interval

        def columns(cols):
            if not isinstance(cols, list):
                raise FilamentError('Columns must be a list, '
                                    '%s found' % type(cols))
            self._cols = cols
            self._tabular = Tabular(self._cols)
            self._tabular.padding_width = 10
            self._tabular.junction_char = '|'

        def add_row(row):
            if not isinstance(row, list):
                raise FilamentError('Expected list type for the row, found %s'
                                    % type(row))
            self._tabular.add_row(row)

        def sort_by(col, sort_desc=True):
            if len(self._cols) == 0:
                raise FilamentError('Expected at least 1 column but 0 found')
            if col not in self._cols:
                raise FilamentError('%s column does not exist' % col)
            self._sort_by = col
            self._sort_desc = sort_desc

        def limit(l):
            if len(self._cols) == 0:
                raise FilamentError('Expected at least 1 column but 0 found')
            if not type(l) is int:
                raise FilamentError('Limit must be an integer value')
            self._limit = l

        def title(text):
            self._tabular.title = text

        self._filament_module.columns = columns
        self._filament_module.title = title
        self._filament_module.sort_by = sort_by
        self._filament_module.limit = limit
        self._filament_module.add_row = add_row
        self._filament_module.render_tabular = self.render_tabular

        on_init = self._find_filament_func('on_init')
        if on_init and self._zero_args(on_init):
            self._filament_module.on_init()
        if self._find_filament_func('on_interval'):
            self.scheduler.add_executor(ThreadPoolExecutor(max_workers=4))
            self.scheduler.start()

            def on_interval():
                try:
                    self._filament_module.on_interval()
                except Exception:
                    self._logger.error('Unexpected error on interval elapsed %s'
                                       % traceback.format_exc())
            self.scheduler.add_job(on_interval,
                                   IntervalTrigger(),
                                   seconds=self._interval,
                                   max_instances=4,
                                   misfire_grace_time=60)
        if len(self._cols) > 0:
            try:
                self._ansi_term.setup_console()
            except TermInitializationError:
                panic('fibratus run: ERROR - console initialization failed')

    def do_output_accessors(self, outputs):
        """Creates the filament's output accessors.

        Parameters
        ----------

        outputs: dict
            outputs initialized from the configuration
            descriptor
        """
        for name, output in outputs.items():
            setattr(self._filament_module, name, OutputAccessor(output))

    def on_next_kevent(self, kevent):
        try:
            self._filament_module.on_next_kevent(ddict(kevent))
        except Exception as e:
            self._logger.error('Unexpected filament error %s' % e)

    def render_tabular(self):
        """Renders the table on the console.
        """
        if len(self._cols) > 0:
            tabular = self._tabular.get_string(start=1, end=self._limit)
            if self._sort_by:
                tabular = self._tabular.get_string(start=1, end=self._limit,
                                                   sortby=self._sort_by,
                                                   reversesort=self._sort_desc)
            self._tabular.clear_rows()
            self._ansi_term.write_output(tabular)

    def close(self):
        on_stop = self._find_filament_func('on_stop')
        if on_stop and self._zero_args(on_stop):
            self._filament_module.on_stop()
        if self.scheduler.running:
            self.scheduler.shutdown()
        self._ansi_term.restore_console()

    @classmethod
    def exists(cls, filament):
        Filament._assert_root_dir()
        return os.path.exists(os.path.join(FILAMENTS_DIR, '%s.py' % filament))

    @classmethod
    def list_filaments(cls):
        Filament._assert_root_dir()
        filaments = {}
        paths = [os.path.join(FILAMENTS_DIR, path) for path in os.listdir(FILAMENTS_DIR)
                 if path.endswith('.py')]
        for path in paths:
            filament_name = os.path.basename(path)[:-3]
            loader = SourceFileLoader(filament_name, path)
            filament = loader.load_module()
            filaments[filament_name] = inspect.getdoc(filament)
        return filaments

    @classmethod
    def _assert_root_dir(cls):
        if not os.path.exists(FILAMENTS_DIR):
            panic('fibratus run: ERROR - %s path does not exist.' % FILAMENTS_DIR)

    @property
    def filters(self):
        return self._filters

    @property
    def logger(self):
        return self._logger

    @logger.setter
    def logger(self, logger):
        self._logger = logger

    @property
    def filament_module(self):
        return self._filament_module

    @property
    def name(self):
        return self._name

    def _find_filament_func(self, func_name):
        """Finds the function in the filament module.

        Parameters
        ----------

        func_name: str
            the name of the function
        """
        functions = inspect.getmembers(self._filament_module, predicate=inspect.isfunction)
        return next(iter([func for name, func in functions if name == func_name]), None)

    def _find_filament_path(self, filament_name):
        """Resolves the filament full path from the name

        Parameters
        ----------

        filament_name: str
            the name of the filament whose path if about to be resolved
        """
        return next(iter([os.path.join(FILAMENTS_DIR, filament) for filament in os.listdir(FILAMENTS_DIR)
                    if filament.endswith('.py') and filament_name == filament[:-3]]), None)

    def _num_args(self, func):
        return len(inspect.getargspec(func).args)

    def _zero_args(self, func):
        return self._num_args(func) == 0



# Copyright 2016 by Nedim Sabic (RabbitStack)
# http://rabbitstack.github.io
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from kstream.includes.windows cimport *

cdef inline SYSTEMTIME sys_time(LARGE_INTEGER timestamp) nogil:
    cdef FILETIME f
    cdef SYSTEMTIME s
    cdef SYSTEMTIME tzt

    f.high_date = timestamp.high
    f.low_date = timestamp.low
    filetime_to_systemtime(&f, &s)
    systemtime_to_tz_specific_localtime(NULL, &s, &tzt)

    return tzt


# Copyright 2017 by Nedim Sabic (RabbitStack)
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from io import TextIOBase

from fibratus.output.fs import FsOutput
from unittest.mock import patch, Mock
import time
import os
import json


class TestFsOutput(object):

    def test_init(self):
        with patch('io.open', return_value=Mock(spec_set=TextIOBase)) as stream_mock:
            fs_output = FsOutput(path='C:\\', mode='a', format='json')
            filename = os.path.join(fs_output.path, '%s.fibra' % time.strftime('%x').replace('/', '-'))
            stream_mock.assert_called_with(filename, 'a')
            assert 'C:\\' in fs_output.path
            assert 'a' in fs_output.mode
            assert 'json' in fs_output.format

    def test_emit(self):
        with patch('io.open', return_value=Mock(spec_set=TextIOBase)):
            fs_output = FsOutput(path='C:\\', mode='a', format='json')
            body = {'kevent_type': 'CreateProcess'}
            fs_output.emit(body)
            fs_output.stream.write.assert_called_with(json.dumps(body) + '\n')


# Copyright 2015 by Nedim Sabic (RabbitStack)
# All Rights Reserved.
# http://rabbitstack.github.io

# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import sys
import re
from prettytable import PrettyTable

__underscore_regex__ = re.compile('((?<=[a-z0-9])[A-Z]|(?!^)[A-Z](?=[a-z]))')


NA = '<NA>'


def panic(msg):
    """Write the message on the console and terminates the process.

    Parameters
    ----------
    msg: str
        the message to be written on the standard output stream
    """
    print(msg)
    sys.exit()


def underscore_dict_keys(in_dict):
    if type(in_dict) is dict:
        out_dict = {}
        for key, item in in_dict.items():
            out_dict[__underscore_regex__.sub(r'_\1', key).lower()] = underscore_dict_keys(item)
        return out_dict
    elif type(in_dict) is list:
        return [__underscore_regex__.sub(r'_\1', obj).lower() for obj in in_dict]
    else:
        return in_dict


class Tabular(PrettyTable):

    def __init__(self, columns, align_col=None, align_type='l', sort_by=None):
        PrettyTable.__init__(self, columns)
        if align_col:
            self.align[align_col] = align_type
        if sort_by:
            self.sortby = sort_by

    def draw(self):
        print(self.get_string())


class DotD(dict):
    """This code is borrowed from easydict
    Credits to:

    https://github.com/makinacorpus/easydict/blob/master/easydict/__init__.py
    """
    def __init__(self, d=None, **kwargs):
        if d is None:
            d = {}
        if kwargs:
            d.update(**kwargs)
        for k, v in d.items():
            setattr(self, k, v)
        # class attributes
        for k in self.__class__.__dict__.keys():
            if not (k.startswith('__') and k.endswith('__')):
                setattr(self, k, getattr(self, k))

    def __setattr__(self, name, value):
        if isinstance(value, (list, tuple)):
            value = [self.__class__(x)
                     if isinstance(x, dict) else x for x in value]
        else:
            value = self.__class__(value) if isinstance(value, dict) else value
        super(DotD, self).__setattr__(name, value)
        super(DotD, self).__setitem__(name, value)

    __setitem__ = __setattr__




# Copyright 2015/2016 by Nedim Sabic (RabbitStack)
# All Rights Reserved.
# http://rabbitstack.github.io

# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
An unusual process attempts to make a network request or it accepts
the incoming connection.
"""

activations = []
processes = ['notepad.exe', 'calc.exe', 'mspaint.exe']


def on_init():
    set_filter('Send', 'Accept', 'Recv', 'Connect')


def on_next_kevent(kevent):
    if kevent.thread:
        process_name = kevent.thread.name
        if process_name in processes:
            triggered = True if process_name in activations else False
            if not triggered:
                message = 'Unusual network activity of kind %s ' \
                          'detected from %s process. ' \
                          'The source ip address is %s and ' \
                          'the destination ip address is %s' \
                           % (kevent.name, process_name,
                              kevent.params.ip_src,
                              kevent.params.ip_dst)
                smtp.emit(message, subject='Anomalous network activity detected')
                activations.append(process_name)



# Copyright 2015 by Nedim Sabic (RabbitStack)
# http://rabbitstack.github.io
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import re
import os
import traceback

from libcpp.unordered_map cimport unordered_map
from cython.operator cimport dereference as deref, preincrement as inc
from libcpp.vector cimport vector
from libcpp.utility cimport pair

from cpython cimport PyBytes_AsString
from cpython.exc cimport PyErr_CheckSignals

from kstream.includes.etw cimport *
from kstream.includes.tdh cimport *
from kstream.includes.windows cimport *
from kstream.includes.python cimport *
from kstream.includes.stdlib cimport *
from kstream.includes.string cimport *
from kstream.time cimport sys_time
from kstream.ktuple cimport build_ktuple
from kstream.process cimport PROCESS_INFO


cdef enum:
    GUID_LENGTH = 36
    INVALID_PID = 4294967295

cdef PyObject* ENUM_PROCESS = build_ktuple(<PyObject*>'{3d6fa8d0-fe05-11d0-9dda-00c04fd7ba7c}', 3)
cdef PyObject* ENUM_THREAD = build_ktuple(<PyObject*>'{3d6fa8d1-fe05-11d0-9dda-00c04fd7ba7c}', 3)
cdef PyObject* ENUM_IMAGE = build_ktuple(<PyObject*>'{2cb15d1d-5fc1-11d2-abe1-00a0c911f518}', 3)
cdef PyObject* REG_CREATE_KCB = build_ktuple(<PyObject*>'{ae53722e-c863-11d2-8659-00c04fa321a1}', 22)
cdef PyObject* REG_DELETE_KCB = build_ktuple(<PyObject*>'{ae53722e-c863-11d2-8659-00c04fa321a1}', 23)

cdef PyObject* CREATE_PROCESS = build_ktuple(<PyObject*>'{3d6fa8d0-fe05-11d0-9dda-00c04fd7ba7c}', 1)
cdef PyObject* TERMINATE_PROCESS = build_ktuple(<PyObject*>'{3d6fa8d0-fe05-11d0-9dda-00c04fd7ba7c}', 2)

cdef wstring PID_PROP = deref_prop("PID")
cdef wstring PPID_PROP = deref_prop("ParentId")
cdef wstring PROCESS_ID_PROP = deref_prop("ProcessId")
cdef wstring IMAGE_FILE_NAME_PROP = deref_prop("ImageFileName")

REGISTRY_KGUID = '{ae53722e-c863-11d2-8659-00c04fa321a1}'
FS_KGUID = '{90cbdc39-4a3e-11d1-84f4-0000f80464e3}'


cdef class KEventStreamCollector:
    """Kernel event stream collector.


    Collects events from the kernel event stream and invokes
    a python callback method for each event delivered
    to the collector.

    The main motivation behind this Cython extension are the perfomance reasons, where
    ETW can generate a huge volume of events, and the parsing
    process is very CPU intensive.

    Use
    ---

    kevt_stream_collector = KEventStreamCollector(logger_name)

    Register a python callback:

    def next_kevt(ktype, cpuid, ts, kparams):
        # your logic here
    kevt_stream_collector.open_kstream(next_kevt)


    """
    cdef EVENT_TRACE_LOGFILE ktrace
    cdef TRACEHANDLE handle
    cdef int pointer_size

    cdef vector[PyObject*]* ktuple_filters
    cdef vector[wchar_t*]* skips
    cdef unordered_map[ULONG, PROCESS_INFO]* proc_map

    cdef ULONG pid_filter
    cdef wchar_t* image_filter
    cdef ULONG own_pid

    cdef next_kevt_callback
    cdef on_kstream_open_callback
    cdef klogger
    cdef regex

    def __init__(self, klogger):
        self.klogger = klogger
        self.handle = 0
        self.next_kevt_callback = None
        self.on_kstream_open_callback = None
        self.ktrace.logger_name = PyBytes_AsString(self.klogger)
        self.regex = re.compile('((?<=[a-z0-9])[A-Z]|(?!^)[A-Z](?=[a-z]))')
        self.pointer_size = 8
        self.ktuple_filters = new vector[PyObject*]()
        self.proc_map = new unordered_map[ULONG, PROCESS_INFO]()
        self.skips = new vector[wchar_t*]()
        self.pid_filter = 0
        self.image_filter = NULL
        self.own_pid = <ULONG>os.getpid()

    def open_kstream(self, callback):
        """Initializes the kernel event stream.

        Sets the event record callback and open
        the trace to consume from kernel event
        stream.

        Parameters
        ----------

        callback: callable
            A python method which is called
            when kernel event is consumed successfully

        """
        self.next_kevt_callback = callback

        self.ktrace.trace_mode = EVENT_TRACE_REAL_TIME_MODE | PROCESS_TRACE_MODE_EVENT_RECORD
        self.ktrace.callback = <PEVENT_RECORD_CALLBACK> self.process_kevent_callback
        # because `process_kevent` callback is the instance
        # method and the ETW API expects a callback function with
        # single parameter, the `self` argument refers to
        # an invalid context. We need to inject the reference
        # to this instance into `Context` member.
        self.ktrace.context = <PVOID> self

        self.handle = open_trace(&self.ktrace)

        if self.on_kstream_open_callback:
            self.on_kstream_open_callback()

        # foward the kernel event stream
        # to the consumer and start the processing
        status = process_trace(&self.handle, 1,
                               NULL,
                               NULL)
        if status != ERROR_SUCCESS or status != ERROR_CANCELLED:
            if status != INVALID_PROCESSTRACE_HANDLE:
                close_trace(self.handle)
            else:
                raise RuntimeError('ERROR - Unable to open kernel event stream. Error %s' % status)

    def close_kstream(self):
        close_trace(self.handle)

    def set_kstream_open_callback(self, callback):
        self.on_kstream_open_callback = callback

    def add_skip(self, skip):
        self.skips.push_back(_wchar_t(<PyObject*>skip))

    def add_ktuple_filter(self, ktuple):
        kguid, opcode = ktuple
        self.ktuple_filters.push_back(build_ktuple(<PyObject*>kguid, <UCHAR>opcode))

    def add_pid_filter(self, pid):
        self.pid_filter = <ULONG>int(pid) if pid else 0

    def add_image_filter(self, image):
        self.image_filter = _wchar_t(<PyObject*>image) if image else NULL

    cdef process_kevent_callback(self, EVENT_RECORD* kevent_trace):
        with nogil:
            (<KEventStreamCollector>kevent_trace.user_ctx)._process_kevent(kevent_trace)

    cdef void _process_kevent(self, EVENT_RECORD* kevent_trace) nogil except *:
        """Kernel event stream callback.

        Parameters
        ----------

        kevent_trace: EVENT_RECORD
            The pointer to kernel event metadata

        """
        cdef TRACE_EVENT_INFO* info = <TRACE_EVENT_INFO*> malloc(4096)

        # the allocation has failed probably
        # because there is no enough memory
        if info == NULL:
            return

        cdef EVENT_HEADER kevt_hdr = kevent_trace.header
        cdef ULONG buffer_size = 4096
        cdef unordered_map[wstring, PyObject*] params
        cdef ULONG property_size
        cdef PROPERTY_DATA_DESCRIPTOR descriptor
        cdef BOOL dropped = False
        cdef PROCESS_INFO pi

        status = tdh_get_event_information(kevent_trace, 0,
                                           NULL,
                                           info,
                                           &buffer_size)

        cpuid = <UCHAR> kevent_trace.buffer_ctx.cpuid
        opcode = <BYTE> kevt_hdr.descriptor.opcode
        pid = <ULONG> kevt_hdr.process_id
        tid = <ULONG> kevt_hdr.thread_id

        ktuple = self.__wrap_ktuple(info.event_guid, opcode)
        # this shouldn't happen, but just in
        # case simply discard the kernel event
        if ktuple == NULL:
            free(info)
            return
        dropped = self.__apply_filters(pid, tid, ktuple, params, True)

        if dropped:
            with gil:
                free(info)
                Py_XDECREF(ktuple)
            return

        if (kevt_hdr.flags & EVENT_HEADER_FLAG_32_BIT_HEADER) == \
                EVENT_HEADER_FLAG_32_BIT_HEADER:
            self.pointer_size = 4
        else:
            self.pointer_size = 8

        if status == ERROR_SUCCESS:
            props = info.properties
            for i from 0 <= i < info.property_count:
                prop = <EVENT_PROPERTY_INFO> props[i]

                property_name = <LPTSTR><BYTE*>info + prop.name_offset

                descriptor.property_name = <ULONGLONG><BYTE*>info + \
                                           prop.name_offset
                descriptor.array_index = 0xFFFFFFFF

                # get the property size which
                # is used to allocate the buffer
                tdh_get_property_size(kevent_trace, 0,
                                      NULL,
                                      1,
                                      &descriptor,
                                      &property_size)
                property_buffer = <BYTE*> malloc(property_size)
                if property_buffer == NULL:
                    return

                # fill the property buffer
                status = tdh_get_property(kevent_trace, 0,
                                          NULL,
                                          1,
                                          &descriptor,
                                          property_size,
                                          property_buffer)
                # get the property value and store it in the map
                if status == ERROR_SUCCESS:
                    if property_name != NULL:
                        mapk = new wstring(<wchar_t*>property_name)
                        params[deref(mapk)] = \
                            self.__parse_property(prop.non_struct_type.in_type,
                                                  prop.non_struct_type.out_type,
                                                  property_buffer)
                        del mapk
                    free(property_buffer)
                else:
                    if property_buffer != NULL:
                        free(property_buffer)

            free(info)
            ts = sys_time(kevt_hdr.timestamp)

            # build a tiny state machine around the
            # currently running processes on the system
            if self.__ktuple_equals(ktuple, ENUM_PROCESS) or \
                self.__ktuple_equals(ktuple, CREATE_PROCESS):
                pi.pid = <ULONG>wcstol(_wchar_t(params.at(PROCESS_ID_PROP)), NULL, 16)
                pi.ppid = <ULONG>wcstol(_wchar_t(params.at(PPID_PROP)), NULL, 16)
                pi.name = _wchar_t(params.at(IMAGE_FILE_NAME_PROP))
                k = new pair[ULONG, PROCESS_INFO](<ULONG>wcstol(_wchar_t(params.at(PROCESS_ID_PROP)), NULL, 16),
                                                 pi)
                self.proc_map.insert(deref(k))
                del k
            elif self.__ktuple_equals(ktuple, TERMINATE_PROCESS):
                # defer the removal of the pid to be able to capture
                # `TerminateProcess` if the image filter is set
                if self.image_filter == NULL:
                    prop_pid = <ULONG>wcstol(_wchar_t(params.at(PROCESS_ID_PROP)), NULL, 16)
                    self.proc_map.erase(prop_pid)

            dropped = self.__apply_filters(pid, tid, ktuple, params, False)
            # now we can erase the pid
            if self.image_filter != NULL and \
                    self.__ktuple_equals(ktuple, TERMINATE_PROCESS):
                prop_pid = <ULONG>wcstol(_wchar_t(params.at(PROCESS_ID_PROP)), NULL, 16)
                self.proc_map.erase(prop_pid)
            if dropped:
                with gil:
                    # decrement references to avoid memory leaks
                    if self.image_filter != NULL:
                       self._decref_params(params)
                    Py_XDECREF(ktuple)
                return
            with gil:
                # check for pending signals.
                # The default behaviour is to
                # raise `KeyboardInterrupt` exception
                # which will be propagated to the caller
                if PyErr_CheckSignals() > 0:
                    self.close_kstream()
                    return
                try:
                    timestamp = '%d-%d-%d %d:%02d:%02d.%d' % (ts.year, ts.month,
                                                              ts.day, ts.hour,
                                                              ts.minute, ts.second,
                                                              ts.millis)
                    # convert the property name from
                    # camel case to underscore so we have
                    # PEP 8 compliant coding style
                    kparams = {self._underscore(self._decref(_wstring(kparam.first))): self._decref(kparam.second)
                               for kparam in params
                               if kparam.second != NULL}
                    kguid, opc = <object>ktuple
                    kguid = kguid[:-1]
                    # registry events have a valid process/thread id
                    # in the event header so we aggregate them
                    if kguid in REGISTRY_KGUID:
                        kparams['thread_id']  = tid
                        kparams['process_id'] = pid
                    elif kguid in FS_KGUID:
                        kparams['process_id']  = pid
                    if self.next_kevt_callback:
                        self.next_kevt_callback((kguid, opc,), cpuid,
                                                timestamp,
                                                kparams)
                except Exception as e:
                    print(traceback.print_exc())
                except KeyboardInterrupt:
                    pass
                finally:
                    Py_XDECREF(ktuple)
        else:
            free(info)

    cdef _decref(self, PyObject* o):
        pyo = None
        if o != NULL:
            pyo = <object>o
            Py_XDECREF(o)
        return pyo

    cdef _decref_params(self,  unordered_map[wstring, PyObject*] params):
        for kparam in params:
            Py_XDECREF(_wstring(kparam.first))
            if kparam.second != NULL:
                Py_XDECREF(kparam.second)

    cdef _underscore(self, o):
        return self.regex.sub(r'_\1', o).lower()

    cdef PyObject* __parse_property(self, USHORT in_type, USHORT out_type,
                                    BYTE* buf) nogil:
        """Parses the property value.

        Given the property input / output types,
        transforms the buffer with property payload.

        Parameters
        ----------

        {in|out}_type : USHORT
            The property input/output types. An input type can
            have multiple output types. For example, UINT16 input
            type can have HEXINT16 or PORT output types.

        buffer: BYTE
            The pointer to the property buffer.
        """
        if buf == NULL:
            return NULL

        if in_type == TDH_INTYPE_UNICODESTRING:
            return _unicode(<wchar_t*>buf)
        elif in_type == TDH_INTYPE_ANSISTRING:
            return _ansi(<char*>buf)
        elif in_type == TDH_INTYPE_UNICODECHAR:
            return _unicodec(buf)
        elif in_type == TDH_INTYPE_ANSICHAR:
            return _ansic(buf)

        elif in_type == TDH_INTYPE_INT8:
            return _i8(buf)
        elif in_type == TDH_INTYPE_UINT8:
            if out_type == TDH_OUTTYPE_HEXINT8:
                return _u8_hex(buf)
            else:
                return _u8(buf)

        elif in_type == TDH_INTYPE_INT16:
            return _i16(buf)
        elif in_type == TDH_INTYPE_UINT16:
            if out_type == TDH_OUTTYPE_HEXINT16:
                return _i16_hex(buf)
            elif out_type == TDH_OUTTYPE_PORT:
                return _ntohs(buf)
            else:
                return _u16(buf)

        elif in_type == TDH_INTYPE_INT32:
            return _i32(buf)
        elif in_type == TDH_INTYPE_UINT32:
            if out_type == TDH_OUTTYPE_HEXINT32:
                return _i32_hex(buf)
            elif out_type == TDH_OUTTYPE_IPV4:
                return ip_addr(buf)
            else:
                return _u32(buf)

        elif in_type == TDH_INTYPE_INT64:
            return _i64(buf)
        elif in_type == TDH_INTYPE_UINT64:
            if out_type == TDH_OUTTYPE_HEXINT64:
                return _i64_hex(buf)
            else:
                return _u64(buf)

        elif in_type == TDH_INTYPE_HEXINT32:
            return _i32_hex(buf)
        elif in_type ==  TDH_INTYPE_HEXINT64:
            return  _i64_hex(buf)

        elif in_type == TDH_INTYPE_FLOAT:
            return _float(buf)
        elif in_type == TDH_INTYPE_DOUBLE:
            return _double(buf)

        elif in_type == TDH_INTYPE_POINTER or \
                        in_type == TDH_INTYPE_SIZET:
            if self.pointer_size == 8:
                return _u64(buf)
            else:
                return _u32(buf)
        else:
            return NULL

    cdef BOOL __apply_filters(self, ULONG pid, ULONG tid,
                              PyObject* ktuple,
                              unordered_map[wstring, PyObject*] params,
                              BOOL defer) nogil:
        cdef BOOL drop = True

        # we don't want to capture any events
        # coming from the fibratus process
        # nor from the process we've declared
        # in the excluded process list
        if self.own_pid == pid:
            return True

        if self.__ktuple_equals(ktuple, ENUM_PROCESS) or \
            self.__ktuple_equals(ktuple, ENUM_THREAD) or \
            self.__ktuple_equals(ktuple, ENUM_IMAGE) or \
            self.__ktuple_equals(ktuple, REG_CREATE_KCB) or \
            self.__ktuple_equals(ktuple, REG_DELETE_KCB):
            return False

        # apply skip list as defined
        # in the configuration descriptor
        drop = self.__apply_skips(pid)
        if drop:
            return True
        elif self.image_filter != NULL:
            drop = True

        for i from 0 <= i < self.ktuple_filters.size():
            ktuple_filter = self.ktuple_filters.at(i)
            if self.__ktuple_equals(ktuple, ktuple_filter):
                drop = False
                break

        # apply pid filter
        if self.pid_filter != 0 and self.pid_filter != pid:
            drop = True
            # we got an invalid pid from the header
            # and we can't still drop the event
            if pid == INVALID_PID:
                drop = False
        elif self.image_filter != NULL:
            # apply image filter
            if defer:
                return False
            if pid == INVALID_PID:
                drop = False
            else:
                drop = self.__apply_image_filter(pid)
                # this only apply to `CreateProcess` events where
                # parent pid is mapped to an image which doesn't match
                # the child pid's image
                if drop and \
                        self.__ktuple_equals(ktuple, CREATE_PROCESS):
                    if params.size() > 0:
                        image_name = _wchar_t(params.at(IMAGE_FILE_NAME_PROP))
                        drop = wcscmp(_wcslwr(image_name),
                                      _wcslwr(self.image_filter)) != 0
        if drop:
            return True
        # now scan for the kernel event
        # properties to find the pid value
        drop = self.__apply_prop_filters(params)

        return drop

    cdef inline BOOL __apply_skips(self, ULONG pid) nogil:
        cdef BOOL ignored = False
        cdef unordered_map[ULONG, PROCESS_INFO].iterator proc_iterator = self.proc_map.find(pid)
        if proc_iterator != self.proc_map.end():
            for i from 0 <= i < self.skips.size():
                skip = self.skips.at(i)
                # compare the image name found
                # on the proc map with the value
                # as defined on the skip list
                pi = deref(proc_iterator).second
                if wcscmp(_wcslwr(pi.name), _wcslwr(skip)) == 0:
                    ignored = True
                    break
        return ignored

    cdef inline BOOL __apply_prop_filters(self, unordered_map[wstring, PyObject*] params) nogil:
        cdef unordered_map[wstring, PyObject*].iterator piter = params.begin()
        cdef BOOL drop = False
        cdef ULONG pid = 0

        while piter != params.end():
            prop = deref(piter)
            prop_name = prop.first

            # get the value of the pid property
            if prop_name.compare(PID_PROP) == 0:
                pid = PyLong_AsLong(prop.second)

            # apply the filters. At this point
            # we also check the kernel event is
            # not coming from the fibratus process
            if pid != 0:
                if pid == self.own_pid:
                    drop = True
                    break
                elif self.pid_filter != 0 and self.pid_filter != pid:
                    drop = True
                    break
                elif self.image_filter != NULL:
                    drop = self.__apply_image_filter(pid)
                    if drop:
                        break
            inc(piter)
        return drop

    cdef inline BOOL __apply_image_filter(self, ULONG pid) nogil:
        cdef BOOL drop = True
        proc_iterator = self.proc_map.find(pid)
        if proc_iterator != self.proc_map.end():
            pi = deref(proc_iterator).second
            drop = wcscmp(_wcslwr(pi.name),
                          _wcslwr(self.image_filter)) != 0
        return drop

    cdef PyObject* __wrap_ktuple(self, GUID guid, UCHAR opcode) nogil:
        cdef wchar_t buf[39]

        if string_from_guid(guid, buf, 39) > 0:
            kguid = PyUnicode_FromWideChar(_wcslwr(buf), 39)
            return build_ktuple(kguid, opcode)
        else:
            return NULL

    cdef inline BOOL __ktuple_equals(self, PyObject* k1, PyObject* k2) nogil:
        cdef BOOL ktuple_equals = False
        if PyLong_AsLong(PyTuple_GetItem(k1, 1)) == PyLong_AsLong(PyTuple_GetItem(k2, 1)):
            kguid1 = _wchar_t(PyTuple_GetItem(k1, 0))
            kguid2 = _wchar_t(PyTuple_GetItem(k2, 0))
            ktuple_equals = wcscmp(kguid1, kguid2) == 0
            if kguid1 != NULL:
                PyMem_Free(kguid1)
            if kguid2 != NULL:
                PyMem_Free(kguid2)
        return ktuple_equals



# Copyright 2015 by Nedim Sabic (RabbitStack)
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from ctypes import Structure, POINTER
from ctypes import c_uint64, c_ulong, c_long, c_ulonglong, c_wchar_p, c_ubyte
from ctypes.wintypes import LARGE_INTEGER, HANDLE

from fibratus.apidefs.guiddef import GUID
import fibratus.apidefs.declarer as declarer


TRACEHANDLE = c_uint64

WNODE_FLAG_TRACED_GUID = 0x00020000
PROCESS_TRACE_MODE_REAL_TIME = 0x00000100


KERNEL_TRACE_CONTROL_GUID = GUID('{9e814aad-3204-11d2-9a82-006008a86939}')
KERNEL_LOGGER_NAME = "NT Kernel Logger"


# enable flags for kernel events
EVENT_TRACE_FLAG_PROCESS = 0x00000001
EVENT_TRACE_FLAG_THREAD = 0x00000002
EVENT_TRACE_FLAG_IMAGE_LOAD = 0x00000004

EVENT_TRACE_FLAG_DISK_IO = 0x00000100
EVENT_TRACE_FLAG_DISK_FILE_IO = 0x00000200

EVENT_TRACE_FLAG_MEMORY_PAGE_FAULTS = 0x00001000
EVENT_TRACE_FLAG_MEMORY_HARD_FAULTS = 0x00002000

EVENT_TRACE_FLAG_NETWORK_TCPIP = 0x00010000

EVENT_TRACE_FLAG_REGISTRY = 0x00020000
EVENT_TRACE_FLAG_DBGPRINT = 0x00040000

EVENT_TRACE_FLAG_PROCESS_COUNTERS = 0x00000008
EVENT_TRACE_FLAG_CSWITCH = 0x00000010
EVENT_TRACE_FLAG_DPC = 0x00000020
EVENT_TRACE_FLAG_INTERRUPT = 0x00000040
EVENT_TRACE_FLAG_SYSTEMCALL = 0x00000080

EVENT_TRACE_FLAG_DISK_IO_INIT = 0x00000400

EVENT_TRACE_FLAG_ALPC = 0x00100000
EVENT_TRACE_FLAG_SPLIT_IO = 0x00200000

EVENT_TRACE_FLAG_DRIVER = 0x00800000
EVENT_TRACE_FLAG_PROFILE = 0x01000000
EVENT_TRACE_FLAG_FILE_IO = 0x02000000
EVENT_TRACE_FLAG_FILE_IO_INIT = 0x04000000


EVENT_TRACE_FLAG_DISPATCHER = 0x00000800
EVENT_TRACE_FLAG_VIRTUAL_ALLOC = 0x00004000


EVENT_TRACE_CONTROL_QUERY = 0
EVENT_TRACE_CONTROL_STOP = 1
EVENT_TRACE_CONTROL_UPDATE = 2


EVENT_CONTROL_CODE_DISABLE_PROVIDER = 0
EVENT_CONTROL_CODE_ENABLE_PROVIDER = 1
EVENT_CONTROL_CODE_CAPTURE_STATE = 2


class WNODE_HEADER(Structure):
    _fields_ = [('buffer_size', c_ulong),
                ('provider_id', c_ulong),
                ('historical_context', c_uint64),
                ('timestamp', LARGE_INTEGER),
                ('guid', GUID),
                ('client_context', c_ulong),
                ('flags', c_ulong)]


class EVENT_TRACE_PROPERTIES(Structure):
    _fields_ = [('wnode', WNODE_HEADER),
                ('buffer_size', c_ulong),
                ('minimum_buffers', c_ulong),
                ('maximum_buffers', c_ulong),
                ('maximum_file_size', c_ulong),
                ('log_file_mode', c_ulong),
                ('flush_timer', c_ulong),
                ('enable_flags', c_ulong),
                ('age_limit', c_long),
                ('number_of_buffers', c_ulong),
                ('free_buffers', c_ulong),
                ('events_lost', c_ulong),
                ('buffers_written', c_ulong),
                ('log_buffers_lost', c_ulong),
                ('real_time_buffer_lost', c_ulong),
                ('logger_thread_id', HANDLE),
                ('log_file_name_offset', c_ulong),
                ('logger_name_offset', c_ulong)]


class TRACE_GUID_REGISTRATION(Structure):
    _fields_ = [('guid', POINTER(GUID)),
                ('reg_handle', HANDLE)]


class EVENT_FILTER_DESCRIPTOR(Structure):
    _fields_ = [('Ptr', c_ulonglong),
                ('Size', c_ulong),
                ('Type', c_ulong)]


class ENABLE_TRACE_PARAMETERS(Structure):
    _fields_ = [('Version', c_ulong),
                ('EnableProperty', c_ulong),
                ('ControlFlags', c_ulong),
                ('SourceId', GUID),
                ('EnableFilterDesc', POINTER(EVENT_FILTER_DESCRIPTOR)),
                ('FilterDescCount', c_ulong)]


start_trace = declarer.declare(declarer.ADVAPI, 'StartTraceW',
                               [POINTER(TRACEHANDLE), c_wchar_p, POINTER(EVENT_TRACE_PROPERTIES)],
                               c_ulong)


control_trace = declarer.declare(declarer.ADVAPI, 'ControlTraceW',
                                 [TRACEHANDLE, c_wchar_p, POINTER(EVENT_TRACE_PROPERTIES), c_ulong],
                                 c_ulong)

enable_trace_ex = declarer.declare(declarer.ADVAPI, 'EnableTraceEx2',
                                   [TRACEHANDLE, POINTER(GUID), c_ulong, c_ubyte, c_ulonglong,
                                    c_ulonglong, c_ulong, POINTER(ENABLE_TRACE_PARAMETERS)],
                                   c_ulong)


# Copyright 2015 by Nedim Sabic (RabbitStack)
# All Rights Reserved.
# http://rabbitstack.github.io

# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from unittest.mock import Mock

import pytest


from fibratus.common import DotD as dd, NA
from fibratus.fs import FsIO, FileOps
from fibratus.handle import HandleInfo, HandleType
from fibratus.kevent import KEvent
from fibratus.kevent_types import CREATE_FILE, DELETE_FILE, WRITE_FILE, RENAME_FILE, SET_FILE_INFORMATION
from fibratus.thread import ThreadRegistry


@pytest.fixture(scope='module')
def kevent():
    return KEvent(Mock(spec_set=ThreadRegistry))


@pytest.fixture(scope='module')
def fsio(kevent):
    handles = [HandleInfo(3080, 18446738026482168384, HandleType.DIRECTORY,
                          "\\Device\\HarddiskVolume2\\Users\\Nedo\\AppData\\Local\\VirtualStore", 640),
               HandleInfo(2010, 18446738023471035392, HandleType.FILE,
                          "\\Device\\HarddiskVolume2\\Windows\\system32\\rpcss.dll", 640)]
    fsio = FsIO(kevent, handles)
    fsio.file_pool[18446738026474426144] = '\\Device\\HarddiskVolume2\\fibratus.log'
    return fsio


class TestFsIO():

    def test_init_fsio(self, fsio):
        assert len(fsio.file_handles) == 2

    @pytest.mark.parametrize('expected_op, kfsio',
                             [(FileOps.SUPERSEDE, dd({"file_object": 18446738026482168384, "ttid": 1484,
                                                      "process_id": 859,
                                                      "create_options": 1223456,
                                         "open_path": "\\Device\\HarddiskVolume2\\Windows\\system32\\kernel32.dll",
                                         "irp_ptr": 18446738026471032392, "share_access": 1, "file_attributes": 0})),
                              (FileOps.OPEN, dd({"file_object": 18446738026482168384, "ttid": 1484, "process_id": 859,
                                                 "create_options": 18874368,
                                         "open_path": "\\Device\\HarddiskVolume2\\Windows\\system32\\kernel32.dll",
                                         "irp_ptr": 18446738026471032392, "share_access": 2, "file_attributes": 0})),
                              (FileOps.CREATE, dd({"file_object": 18446738026482168384, "ttid": 1484, "process_id": 859,
                                                   "create_options": 33554532,
                                         "open_path": "\\Device\\HarddiskVolume2\\Windows\\system32\\kernel32.dll",
                                         "irp_ptr": 18446738026471032392, "share_access": 4, "file_attributes": 0})),
                              (FileOps.OPEN_IF, dd({"file_object": 18446738026482168384, "ttid": 1484,
                                                    "process_id": 859,
                                                    "create_options": 58651617,
                                         "open_path": "\\Device\\HarddiskVolume2\\Windows\\system32\\kernel32.dll",
                                         "irp_ptr": 18446738026471032392, "share_access": 3, "file_attributes": 0})),
                              (FileOps.OVERWRITE, dd({"file_object": 18446738026482168384, "ttid": 1484,
                                                      "process_id": 859,
                                                      "create_options": 78874400,
                                         "open_path": "\\Device\\HarddiskVolume2\\Windows\\system32\\kernel32.dll",
                                         "irp_ptr": 18446738026471032392, "share_access": 5, "file_attributes": 0})),
                              (FileOps.OVERWRITE_IF, dd({"file_object": 18446738026482168384, "ttid": 1484,
                                                         "process_id": 859,
                                                         "create_options": 83886112,
                                         "open_path": "\\Device\\HarddiskVolume2\\Windows\\system32\\kernel32.dll",
                                         "irp_ptr": 18446738026471032392, "share_access": 6, "file_attributes": 0}))])
    def test_create_file_operation(self, expected_op, kfsio, fsio, kevent):

        fsio.parse_fsio(CREATE_FILE, kfsio)

        kparams = kevent.params
        assert kparams.file == kfsio.open_path
        assert kparams.tid == kfsio.ttid
        assert kparams.pid == kfsio.process_id
        assert kparams.operation == expected_op.name

    @pytest.mark.parametrize('expected_share_mask, kfsio',
                             [('r--', dd({"file_object": 18446738026482168384, "ttid": 1484, "create_options": 18874368,
                                         "open_path": "\\Device\\HarddiskVolume2\\Windows\\system32\\kernel32.dll",
                                         "process_id": 859,
                                         "irp_ptr": 18446738026471032392, "share_access": 1, "file_attributes": 0})),
                              ('-w-', dd({"file_object": 18446738026482168384, "ttid": 1484, "create_options": 18874368,
                                         "open_path": "\\Device\\HarddiskVolume2\\Windows\\system32\\kernel32.dll",
                                          "process_id": 859,
                                         "irp_ptr": 18446738026471032392, "share_access": 2, "file_attributes": 0})),
                              ('--d', dd({"file_object": 18446738026482168384, "ttid": 1484, "create_options": 18874368,
                                         "open_path": "\\Device\\HarddiskVolume2\\Windows\\system32\\kernel32.dll",
                                         "process_id": 859,
                                         "irp_ptr": 18446738026471032392, "share_access": 4, "file_attributes": 0})),
                              ('rw-', dd({"file_object": 18446738026482168384, "ttid": 1484, "create_options": 18874368,
                                         "open_path": "\\Device\\HarddiskVolume2\\Windows\\system32\\kernel32.dll",
                                         "process_id": 859,
                                         "irp_ptr": 18446738026471032392, "share_access": 3, "file_attributes": 0})),
                              ('r-d', dd({"file_object": 18446738026482168384, "ttid": 1484, "create_options": 18874368,
                                         "open_path": "\\Device\\HarddiskVolume2\\Windows\\system32\\kernel32.dll",
                                         "process_id": 859,
                                         "irp_ptr": 18446738026471032392, "share_access": 5, "file_attributes": 0})),
                              ('-wd', dd({"file_object": 18446738026482168384, "ttid": 1484, "create_options": 18874368,
                                         "open_path": "\\Device\\HarddiskVolume2\\Windows\\system32\\kernel32.dll",
                                         "process_id": 859,
                                         "irp_ptr": 18446738026471032392, "share_access": 6, "file_attributes": 0})),
                              ('rwd', dd({"file_object": 18446738026482168384, "ttid": 1484, "create_options": 18874368,
                                         "open_path": "\\Device\\HarddiskVolume2\\Windows\\system32\\kernel32.dll",
                                         "process_id": 859,
                                         "irp_ptr": 18446738026471032392, "share_access": 7, "file_attributes": 0})),
                              ('---', dd({"file_object": 18446738026482168384, "ttid": 1484, "create_options": 18874368,
                                         "open_path": "\\Device\\HarddiskVolume2\\Windows\\system32\\kernel32.dll",
                                         "process_id": 859,
                                         "irp_ptr": 18446738026471032392, "share_access": -1, "file_attributes": 0}))])
    def test_create_file_share_mask(self, expected_share_mask, kfsio, fsio, kevent):
        fsio.parse_fsio(CREATE_FILE, kfsio)
        assert kevent.params.share_mask == expected_share_mask

    def test_delete_file(self, fsio, kevent):
        kfsio = dd({"file_object": 18446738026474426144, "ttid": 1956, "process_id": 859, "irp_ptr": 18446738026471032392})
        fsio.parse_fsio(DELETE_FILE, kfsio)
        assert kevent.params.tid == kfsio.ttid
        assert kevent.params.file == '\\Device\\HarddiskVolume2\\fibratus.log'

    def test_write_file(self, fsio, kevent):
        kfsio = dd({"file_object": 18446738026474426144, "process_id": 859, "io_flags": 0, "io_size": 8296,
                    "offset": 75279, "ttid": 1956})
        fsio.parse_fsio(WRITE_FILE, kfsio)
        assert kevent.params.tid == kfsio.ttid
        assert kevent.params.file == NA
        assert kevent.params.io_size == kfsio.io_size / 1024

    def test_rename_file(self, fsio, kevent):
        kfsio = dd({"file_object": 18446738023471035392, "ttid": 1956, "process_id": 859, "irp_ptr": 18446738026471032392})
        fsio.parse_fsio(RENAME_FILE, kfsio)
        assert kevent.params.tid == kfsio.ttid
        assert kevent.params.file == '\\Device\\HarddiskVolume2\\Windows\\system32\\rpcss.dll'

    def test_set_file_information(self, fsio, kevent):
        kfsio = dd(
            {"file_object": 18446738023471035392, "ttid": 1956, "info_class": 20, "process_id": 859,
             "irp_ptr": 18446738026471032392})
        fsio.parse_fsio(SET_FILE_INFORMATION, kfsio)
        assert kevent.params.tid == kfsio.ttid
        assert kevent.params.info_class == 20
        assert kevent.params.file == '\\Device\\HarddiskVolume2\\Windows\\system32\\rpcss.dll'



# Copyright 2016 by Nedim Sabic (RabbitStack)
# All Rights Reserved.
# http://rabbitstack.github.io

# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
Shows top processes by registry I/O activity.
"""

import collections

processes_registry_io = collections.Counter()


def on_init():
    set_filter('RegOpenKey', 'RegQueryKey', 'RegCreateKey', 'RegQueryValue', 'RegSetValue', 'RegDeleteValue')
    columns(["Process", "#Ops"])
    sort_by('#Ops')
    set_interval(1)
    limit(20)


def on_next_kevent(kevent):
    process = ['%s (%d)' % (kevent.thread.name, kevent.thread.pid)]
    processes_registry_io.update(process)


def on_interval():
    for process, io in processes_registry_io.items():
        add_row([process, io])
    render_tabular()



# Copyright 2017 by Nedim Sabic (RabbitStack)
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from fibratus.binding.base import BaseBinding
from fibratus.errors import BindingError

import os
import glob

try:
    import yara
except ImportError:
    yara = None


class YaraBinding(BaseBinding):

    def __init__(self, outputs, logger, **config):
        """Creates an instance of the YARA binding.

        This binding integrates with YARA tool to provide real time classification and pattern matching of the
        process's binary images. The image path is extracted from the `ThreadInfo` instance after `CreateProcess`
        kernel event has been captured.

        :param dict outputs: declared output adapters
        :param logbook.Logger logger: reference to the logger implementation
        :param dict config: configuration for this binding
        """

        BaseBinding.__init__(self, outputs, logger)
        self._path = config.pop('path', None)
        self._rules = None
        if not yara:
            raise BindingError('yara-python package is not installed')
        if not os.path.exists(self._path) or not os.path.isdir(self._path):
            raise BindingError('%s rules path does not exist' %
                               self._path)
        try:
            for file in glob.glob(os.path.join(self._path, '*.yar')):
                self._rules = yara.compile(os.path.join(self._path, file))
        except yara.SyntaxError as e:
            raise BindingError("rule compilation error %s" % e)

    def run(self, **kwargs):
        """Apply the YARA rule set to process's image path.

        If a rule match occurs, the data with rule information, matching strings, process name, etc. is transported
        over provided output implementation. If output type is not specified, the console output stream is used.

        :param dict kwargs: parameters for the binding context
        """
        thread_info = kwargs.pop('thread_info', None)
        kevent = kwargs.pop('kevent', None)
        if thread_info:
            def yara_callback(data):
                matches = data['matches']
                if matches:
                    rule_context = {
                        'rule_info': {
                            'meta': data['meta'],
                            'tags': data['tags'],
                            'namespace': data['namespace'],
                            'rule': data['rule'],
                            'strings': [self.__string_meta(string) for string in data['strings']]
                        }
                    }
                    kevent.params.update(rule_context)
                return yara.CALLBACK_CONTINUE
            self._rules.match(thread_info.exe, callback=yara_callback)

    def __string_meta(self, string):
        """Unpacks the tuple with matching string data and transforms it to a dictionary.

        :param tuple string: the tuple with matching string data
        :return: dict:
        """
        offset, ident, data = string
        return {
            'offset': offset,
            'identifier': ident,
            'data': data.decode('utf-8')
        }



# Copyright 2015 by Nedim Sabic (RabbitStack)
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from enum import Enum

from fibratus.common import NA
from fibratus.handle import HandleType
from fibratus.kevent_types import *
from fibratus.apidefs.fs import *


class FileOps(Enum):

    # if the file already exists,
    # replace it with the given file
    # otherwise create the given file
    SUPERSEDE = 0
    # if the file already exists,
    # open it instead of creating a new file
    OPEN = 1
    # if the file already exists,
    # fail the request otherwise create the file
    CREATE = 2
    # if the file already exists,
    # open it, otherwise create the given file
    OPEN_IF = 3
    # if the file already exists,
    # open it and overwrite it,
    # otherwise fail the request
    OVERWRITE = 4
    # if the file already exists,
    # open it and overwrite it,
    # otherwise create the given file
    OVERWRITE_IF = 5


class FileType(Enum):

    FILE = 0
    DIRECTORY = 1
    REPARSE_POINT = 2
    UNKNOWN = 3


class FsIO(object):

    def __init__(self, kevent, handles):
        self._kevent = kevent
        self.file_pool = {}
        self.file_handles = {handle.obj: (handle.name, handle.handle_type, handle.handle)
                             for handle in handles
                             if handle.handle_type in [HandleType.FILE, HandleType.DIRECTORY]}

    def parse_fsio(self, ketype, kfsio):
        """Parses the file system related kevents.

        Parameters
        ----------

        ketype: tuple
            kevent type
        kfsio: dict
            kevent payload as forwarded from
        """

        # thread which is perfoming the op
        tid = kfsio.ttid
        pid = kfsio.process_id
        obj = kfsio.file_object
        self._kevent.tid = tid
        self._kevent.pid = pid
        # creates or opens a file or the I/O device.
        # The device can be a file, file stream, directory,
        # physical disk, volume, console buffer, tape drive,
        # communications resource, mailslot, or pipe.
        if ketype == CREATE_FILE:
            file = kfsio.open_path
            # the high 8 bits correspond to the value of the
            # `CreateDisposition` parameter and the low 24 bits
            # are the value of the `CreateOptions` parameter
            # of the `NtCreateFile` system call
            co = kfsio.create_options

            # extract the most significat 8 bits
            flags = (co >> 24) & ((1 << 8) - 1)

            op = FileOps.OPEN
            if flags == FILE_SUPERSEDE:
                op = FileOps.SUPERSEDE
            elif flags == FILE_OPEN:
                op = FileOps.OPEN
            elif flags == FILE_CREATE:
                op = FileOps.CREATE
            elif flags == FILE_OPEN_IF:
                op = FileOps.OPEN_IF
            elif flags == FILE_OVERWRITE:
                op = FileOps.OVERWRITE
            elif flags == FILE_OVERWRITE_IF:
                op = FileOps.OVERWRITE_IF

            # determine file descriptor type
            file_type = FileType.FILE
            if (co & FILE_DIRECTORY_FILE) == FILE_DIRECTORY_FILE:
                file_type = FileType.DIRECTORY
            elif (co & FILE_OPEN_REPARSE_POINT) == FILE_OPEN_REPARSE_POINT:
                file_type = FileType.REPARSE_POINT

            share_mask = self._resolve_share_mask(kfsio.share_access)
            params = {
                'file': file,
                'file_type': file_type.name,
                'file_object': obj,
                'tid': tid,
                'pid': pid,
                'operation': op.name,
                'share_mask': share_mask
            }
            self._kevent.params = params

            # index by file object pointer
            # so we can query the pool
            # to resolve the file name
            self.file_pool[obj] = file

        elif ketype == DELETE_FILE or ketype == CLOSE_FILE:
            file = self._query_file_name(obj, True)
            params = {
                'file': file,
                'file_object': obj,
                'pid': pid,
                'tid': tid
            }
            self._kevent.params = params
        elif ketype == WRITE_FILE or ketype == READ_FILE:
            # the number of kb read/written
            io_size = kfsio.io_size / 1024
            file = self._query_file_name(obj)
            params = {
                'file': file,
                'file_object': obj,
                'pid': pid,
                'tid': tid,
                'io_size': io_size
            }
            self._kevent.params = params
        elif ketype == RENAME_FILE:
            file = self._query_file_name(obj)
            params = {
                'file': file,
                'file_object': obj,
                'pid': pid,
                'tid': tid
            }
            self._kevent.params = params
            if NA not in file:
                self.file_pool[obj] = file
        elif ketype == SET_FILE_INFORMATION:
            file = self._query_file_name(obj)
            params = {
                'file': file,
                'file_object': obj,
                'pid': pid,
                'tid': tid,
                'info_class': kfsio.info_class
            }
            self._kevent.params = params

    def _query_file_name(self, fobj, remove=False):
        if fobj in self.file_pool:
            return self.file_pool.pop(fobj) if remove \
                else self.file_pool[fobj]
        else:
            # couldn't find the file in the file pool,
            # query the file handles
            if fobj in self.file_handles:
                file, _, _ = self.file_handles.pop(fobj)
                if file and not remove:
                    self.file_pool[fobj] = file
                return file if file else NA
            else:
                return NA

    def _resolve_share_mask(self, share_access):
        """Resolves the share mask.

         Resolves the type of share access that
         the caller would like to use in the file.

         For example, `FILE_SHARE_READ` would allow other
         threads to open the file for read access.

        :param str share_access: the value of the share access
        :return: str: resolved share mask
        """

        if share_access == FILE_SHARE_READ:
            return 'r--'
        elif share_access == FILE_SHARE_WRITE:
            return '-w-'
        elif share_access == FILE_SHARE_DELETE:
            return '--d'
        elif share_access == (FILE_SHARE_READ | FILE_SHARE_WRITE):
            return 'rw-'
        elif share_access == (FILE_SHARE_READ | FILE_SHARE_DELETE):
            return 'r-d'
        elif share_access == (FILE_SHARE_WRITE | FILE_SHARE_DELETE):
            return '-wd'
        elif share_access == (FILE_SHARE_READ | FILE_SHARE_WRITE |
                              FILE_SHARE_DELETE):
            return 'rwd'
        else:
            return '---'



# Copyright 2016 by Nedim Sabic (RabbitStack)
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from fibratus.output.console import ConsoleOutput


class OutputAggregator(object):

    def __init__(self, outputs):
        self.outputs = outputs

    def aggregate(self, kevent):
        """Emit the kernel stream via output sinks.

        For each output registered, invokes the `emit``
        method to send the kernel event info to the
        output sink.

        Parameters
        ----------

        kevent: KEvent
            an instance of the kernel event
        """
        for _, output in self.outputs.items():
            if isinstance(output, ConsoleOutput):
                output.emit(kevent)
            else:
                pid, proc = kevent.get_thread()
                body = {'id': kevent.kid,
                        'timestamp': kevent.ts.strftime('%Y-%m-%d %H:%M:%S.%f'),
                        'cpuid': kevent.cpuid,
                        'proc': proc,
                        'pid': pid,
                        'name': kevent.name,
                        'category': kevent.category,
                        'params': kevent.params}
                output.emit(body)



# Copyright 2015 by Nedim Sabic (RabbitStack)
# All Rights Reserved.
# http://rabbitstack.github.io

# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from ctypes import POINTER
from ctypes import c_int, c_byte, WINFUNCTYPE
from ctypes.wintypes import DWORD, ULONG, PULONG, USHORT, HANDLE, BOOL, SHORT, WCHAR, CHAR, WORD, LPDWORD

from fibratus.apidefs.cdefs import *
import fibratus.apidefs.declarer as declarer


SYSTEM_HANDLE_INFORMATION_CLASS = 16

PUBLIC_OBJECT_BASIC_INFORMATION = 0
PUBLIC_OBJECT_NAME_INFORMATION = 1
PUBLIC_OBJECT_TYPE_INFORMATION = 2

# this constants may vary from
# one Windows version to another
# for Win 7/8 they should have
# the following values
FILE_OBJECT_TYPE_INDEX = 28


STD_OUTPUT_HANDLE = -11
INVALID_HANDLE_VALUE = -1

CONSOLE_TEXTMODE_BUFFER = 1

GENERIC_READ = 0x80000000
GENERIC_WRITE = 0x40000000

FILE_SHARE_READ = 0x00000001
FILE_SHARE_WRITE = 0x00000002


# console structures
class CURSOR_INFO(ctypes.Structure):
    _fields_ = [("size", c_int),
                ("visible", c_byte)]


class COORD(ctypes.Structure):
    _fields_ = [("x", SHORT), ("y", SHORT)]


class SMALL_RECT(ctypes.Structure):
    _fields_ = [("left", SHORT),
                ("top", SHORT),
                ("right", SHORT),
                ("bottom", SHORT)]


class CHAR_INFOU(ctypes.Union):
    _fields_ = [("unicode_char", WCHAR), ("ascii_char", CHAR)]


class CHAR_INFO(ctypes.Structure):
    _anonymous_ = ("char",)
    _fields_ = [("char", CHAR_INFOU), ("attributes", WORD)]


class CONSOLE_SCREEN_BUFFER_INFO(ctypes.Structure):
    _fields_ = [('size', COORD),
                ('cursor_position', COORD),
                ('attributes', WORD),
                ('window', SMALL_RECT),
                ('maximum_window_size', COORD)]


class SYSTEM_HANDLE(Structure):
    _fields_ = [('process_id', ULONG),
                ('object_type_number', UCHAR),
                ('flags', UCHAR),
                ('handle', USHORT),
                ('object', PVOID),
                ('access_mask', DWORD)]


class SYSTEM_HANDLE_INFORMATION(Structure):
    _fields_ = [('number_of_handles', ULONG),
                ('handles', SYSTEM_HANDLE * 1)]


class OBJECT_TYPE_INFORMATION(Structure):
    _fields_ = [('type_name', UNICODE_STRING),
                ('reserved', ULONG * 22)]


# retrieves the specified system information
zw_query_system_information = declarer.declare(declarer.NT, 'ZwQuerySystemInformation',
                                               [DWORD, PVOID, ULONG, PULONG],
                                               DWORD)

# memory alloc/free functions
malloc = declarer.declare(declarer.C, 'malloc', [c_size_t], c_void_p)
realloc = declarer.declare(declarer.C, 'realloc', [c_void_p, c_size_t], c_void_p)
free = declarer.declare(declarer.C, 'free', [c_void_p], None)

# object handle cleanup
close_handle = declarer.declare(declarer.KERNEL, 'CloseHandle', [HANDLE], BOOL)
# duplicate object handle
duplicate_handle = declarer.declare(declarer.KERNEL, 'DuplicateHandle',
                                    [HANDLE, HANDLE, HANDLE, POINTER(HANDLE), DWORD, ULONG, ULONG],
                                    DWORD)

# query object name / type
nt_query_object = declarer.declare(declarer.NT, 'NtQueryObject',
                                   [HANDLE, ULONG, PVOID, ULONG, PULONG],
                                   DWORD)

# low level console api
get_std_handle = declarer.declare(declarer.KERNEL, 'GetStdHandle', [DWORD], HANDLE)
set_console_active_screen_buffer = declarer.declare(declarer.KERNEL, 'SetConsoleActiveScreenBuffer', [HANDLE], BOOL)

create_console_screen_buffer = declarer.declare(declarer.KERNEL, 'CreateConsoleScreenBuffer',
                                                [DWORD, DWORD, c_void_p, DWORD, LPVOID], HANDLE)
get_console_screen_buffer_info = declarer.declare(declarer.KERNEL, 'GetConsoleScreenBufferInfo',
                                                  [HANDLE, POINTER(CONSOLE_SCREEN_BUFFER_INFO)], BOOL)

write_console_output = declarer.declare(declarer.KERNEL, 'WriteConsoleOutputW',
                                        [HANDLE, POINTER(CHAR_INFO), COORD, COORD, POINTER(SMALL_RECT)], BOOL)

set_console_cursor_position = declarer.declare(declarer.KERNEL, 'SetConsoleCursorPosition',
                                               [HANDLE, COORD], BOOL)

get_console_cursor_info = declarer.declare(declarer.KERNEL, 'GetConsoleCursorInfo',
                                           [HANDLE, POINTER(CURSOR_INFO)], BOOL)

set_console_cursor_info = declarer.declare(declarer.KERNEL, 'SetConsoleCursorInfo',
                                           [HANDLE, POINTER(CURSOR_INFO)], BOOL)

write_console_unicode = declarer.declare(declarer.KERNEL, 'WriteConsoleW',
                                         [HANDLE, c_void_p, DWORD, LPDWORD, LPVOID], BOOL)


PHANDLER_ROUTINE = WINFUNCTYPE(BOOL, DWORD)
set_console_ctrl_handler = declarer.declare(declarer.KERNEL, 'SetConsoleCtrlHandler',
                                            [PHANDLER_ROUTINE, BOOL], BOOL)


# event objects
create_event = declarer.declare(declarer.KERNEL, 'CreateEventW', [c_void_p, BOOL, BOOL, LPTSTR], HANDLE)





# Copyright 2016 by Nedim Sabic (RabbitStack)
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import json
from unittest.mock import patch

import pika
import pytest

from fibratus.errors import InvalidPayloadError
from fibratus.output.amqp import AmqpOutput


@pytest.fixture(scope='module')
def amqp_adapter():
    config = {
        'username': 'fibratus',
        'host': '127.0.0.1',
        'port': 5672,
        'vhost': '/',
        'exchange': 'test',
        'routingkey': 'fibratus'
    }
    return AmqpOutput(**config)


class TestAmqpOutput(object):

    def test_init(self, amqp_adapter):
        assert 'fibratus' in amqp_adapter.username
        assert 'guest' in amqp_adapter._password
        assert '127.0.0.1' in amqp_adapter.host
        assert amqp_adapter.port == 5672
        assert '/' in amqp_adapter.vhost
        assert 'test' in amqp_adapter.exchange
        assert 'fibratus' in amqp_adapter.routingkey
        assert amqp_adapter.delivery_mode == 1

    @patch('pika.BlockingConnection', spec_set=pika.BlockingConnection)
    def test_emit(self, connection_mock, amqp_adapter):
        body = {'kevent_type': 'CreateProcess'}
        amqp_adapter.emit(body)
        connection_mock.channel.assert_called_once()
        amqp_adapter._channel.basic_publish.assert_called_with('test', 'fibratus',
                                                               json.dumps(body),
                                                               amqp_adapter._basic_props)

    @patch('pika.BlockingConnection', spec_set=pika.BlockingConnection)
    def test_emit_invalid_payload(self, connection_mock, amqp_adapter):
        body = ['CrateProcess', 'TerminateProcess']
        with pytest.raises(InvalidPayloadError) as e:
            connection_mock.channel.assert_called_once()
            amqp_adapter.emit(body)
        assert "invalid payload for AMQP message. dict expected but <class 'list'> found" == str(e.value)
        amqp_adapter._channel.basic_publish.assert_not_called()

    @patch('pika.BlockingConnection', spec_set=pika.BlockingConnection)
    def test_emit_override_exchange_and_rk(self, connection_mock, amqp_adapter):
        body = {'kevent_type': 'CreateProcess'}
        amqp_adapter.emit(body, exchange='test.override', routingkey='fibratus.override')
        amqp_adapter._channel.basic_publish.assert_called_with('test.override', 'fibratus.override',
                                                               json.dumps(body),
                                                               amqp_adapter._basic_props)

    @pytest.mark.parametrize('body', [{'kevent_type': 'CreateProcess'}, {'kevent_type': 'TerminateProcess'},
                                      {'kevent_type': 'WriteFile'}, {'kevent_type': 'Recv'}])
    @patch('pika.BlockingConnection', spec_set=pika.BlockingConnection)
    def test_emit_multiple(self, connection_mock, body, amqp_adapter):
        amqp_adapter.emit(body, exchange='test.override', routingkey='fibratus.override')
        connection_mock.channel.assert_called_once()
        amqp_adapter._channel.basic_publish.assert_called_with('test.override', 'fibratus.override',
                                                               json.dumps(body),
                                                               amqp_adapter._basic_props)



# Copyright 2015 by Nedim Sabic (RabbitStack)
# All Rights Reserved.
# http://rabbitstack.github.io

# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from unittest.mock import Mock, patch

import pytest
from fibratus.apidefs.process import THREAD_QUERY_INFORMATION
from fibratus.common import DotD

from fibratus.kevent import KEvent, KEvents, Category
from fibratus.thread import ThreadRegistry, ThreadInfo


@pytest.fixture(scope='module')
def thread_registry_mock():
    thread_registry = Mock(spec_set=ThreadRegistry)
    thread_registry.get_thread.side_effect = [ThreadInfo(436, 1024, 23, 'svchost.exe',
                                                         'C:\\Windows\\system32\\svchost.exe -k RPCSS'),
                                              None,
                                              None,
                                              ThreadInfo(836, 564, 23, 'svchost.exe',
                                                         'C:\\Windows\\system32\\svchost.exe -k RPCSS'),
                                              None,
                                              ThreadInfo(836, 564, 23, 'svchost.exe',
                                                         'C:\\Windows\\system32\\svchost.exe -k RPCSS'),
                                              None, None]
    return thread_registry


@pytest.fixture()
def kevent(thread_registry_mock):
    return KEvent(thread_registry_mock)


class TestKEvent(object):

    def test_get_thread_pid_not_none(self, kevent, thread_registry_mock):
        kevent.pid = 436
        assert kevent.thread
        thread_registry_mock.get_thread.assert_called_with(kevent.pid)

    def test_get_thread_pid_not_none_and_not_found_in_registry(self, kevent, thread_registry_mock):
        kevent.pid = 436
        assert kevent.thread is None
        thread_registry_mock.get_thread.assert_called_with(kevent.pid)

    def test_get_thread_pid_not_none_find_by_thread(self, kevent, thread_registry_mock):
        kevent.pid = 436
        kevent.tid = 564
        assert kevent.thread
        thread_registry_mock.get_thread.assert_called_with(kevent.tid)

    def test_get_thread_pid_none(self, kevent, thread_registry_mock):
        kevent.tid = 564
        assert kevent.thread
        thread_registry_mock.get_thread.assert_called_with(kevent.tid)

    @patch('fibratus.kevent.open_thread', return_value=25)
    @patch('fibratus.kevent.get_process_id_of_thread')
    @patch('fibratus.kevent.close_handle')
    def test_get_thread_pid_none(self, close_handle_mock, get_process_id_of_thread_mock, open_thread_mock,
                                 kevent):
        kevent.tid = 245
        kevent.get_thread()
        open_thread_mock.assert_called_with(THREAD_QUERY_INFORMATION,
                                            False, kevent.tid)
        get_process_id_of_thread_mock.assert_called_with(25)
        close_handle_mock.assert_called_with(25)

    def test_kevents_all(self):
        kevents = KEvents.all()
        assert isinstance(kevents, list)
        assert len(kevents) > 0

    def test_kevents_meta_info(self):
        kevents_meta_info = KEvents.meta_info()
        assert isinstance(kevents_meta_info, dict)
        cat, description = kevents_meta_info[KEvents.CREATE_PROCESS]
        assert cat == Category.PROCESS
        assert description

    def test_set_kevent_name(self, kevent):
        kevent.name = KEvents.CREATE_PROCESS
        assert kevent.name == KEvents.CREATE_PROCESS
        assert kevent.category == Category.PROCESS.name

    def test_set_kevent_ts(self, kevent):
        kevent.ts = '2016-11-29 12:31:48.210000'
        assert kevent.ts.second == 48
        assert kevent.ts.minute == 31
        assert kevent.ts.hour == 12

    def test_set_kevent_cpu_id(self, kevent):
        kevent.cpuid = 2
        assert kevent.cpuid == 2

    def test_set_kevent_params(self, kevent):
        kevent.params = {'exe': 'svchost.exe', 'comm': 'C:\\Windows\\system32\\svchost.exe -k RPCSS'}
        assert isinstance(kevent.params, DotD)
        assert kevent.params.exe == 'svchost.exe'




# Copyright 2015 by Nedim Sabic (RabbitStack)
# All Rights Reserved.
# http://rabbitstack.github.io
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from ctypes import addressof, byref, cast, memmove, sizeof,  c_char, c_wchar
from ctypes import ArgumentError, pointer

from fibratus.apidefs.cdefs import ERROR_ALREADY_EXISTS, ERROR_ACCESS_DENIED, ERROR_BAD_LENGTH, \
    ERROR_INVALID_PARAMETER, ERROR_SUCCESS
from fibratus.apidefs.etw import *
from fibratus.errors import FibratusError
from fibratus.common import panic

class KTraceProps(object):

    def __init__(self, buffer_size=1024):
        """Builds the tracing session properties.

        Parameters
        ---------

        buffer_size: int
            the amount of memory allocated for each trace buffer
        """

        # allocate buffer for the trace
        self.max_string_len = 1024
        self.buff_size = sizeof(EVENT_TRACE_PROPERTIES) + 2 * sizeof(c_wchar) * self.max_string_len

        self._buff = (c_char * self.buff_size)()
        self._props = cast(pointer(self._buff), POINTER(EVENT_TRACE_PROPERTIES))

        # set trace properties
        self._props.contents.wnode.buffer_size = self.buff_size
        self._props.contents.wnode.guid = KERNEL_TRACE_CONTROL_GUID
        self._props.contents.wnode.flags = WNODE_FLAG_TRACED_GUID
        self._props.contents.logger_name_offset = sizeof(EVENT_TRACE_PROPERTIES)
        self._props.contents.log_file_name_offset = 0
        self._props.contents.log_file_mode = PROCESS_TRACE_MODE_REAL_TIME
        self._props.contents.buffer_size = buffer_size

    def enable_kflags(self, syscall=False, cswitch=False):
        # enable the basic set of flags
        # for the kernel events
        self._props.contents.enable_flags = (EVENT_TRACE_FLAG_PROCESS |
                                             EVENT_TRACE_FLAG_REGISTRY |
                                             EVENT_TRACE_FLAG_THREAD |
                                             EVENT_TRACE_FLAG_DISK_IO |
                                             EVENT_TRACE_FLAG_DISK_FILE_IO |
                                             EVENT_TRACE_FLAG_FILE_IO |
                                             EVENT_TRACE_FLAG_FILE_IO_INIT |
                                             EVENT_TRACE_FLAG_IMAGE_LOAD |
                                             EVENT_TRACE_FLAG_NETWORK_TCPIP)

        # syscall / cswitch flags generate a LOT of kevents
        # and they are disabled by default
        if syscall:
            self._props.contents.enable_flags |= (EVENT_TRACE_FLAG_SYSTEMCALL | EVENT_TRACE_FLAG_CSWITCH)
        if cswitch:
            self._props.contents.enable_flags |= EVENT_TRACE_FLAG_CSWITCH

    def get(self):
        return self._props

    @property
    def logger_name(self):
        return c_wchar_p(addressof(self._props.contents) +
                         self._props.contents.logger_name_offset)

    @logger_name.setter
    def logger_name(self, logger_name):
        name_len = len(logger_name) + 1
        if self.max_string_len < name_len:
            raise ArgumentError("Logger name %s is too long" % logger_name)
        props = self._props
        logger = c_wchar_p(addressof(props.contents) + props.contents.logger_name_offset)
        memmove(logger,  c_wchar_p(logger_name), sizeof(c_wchar) * name_len)


class KTraceController(object):
    """Controls the life cycle of the kernel traces.

    """

    def __init__(self):
        self._handle = TRACEHANDLE()
        self._trace_name = None

    def __del__(self):
        if self._handle:
            self.stop_ktrace()

    def start_ktrace(self, name, kprops):
        """Starts a new trace.

        Parameters
        ---------

        name: str
            the name for the trace session
        kprops: KTraceProps
            an instance of the kernel trace properties
        """
        self._trace_name = name
        handle = TRACEHANDLE()
        kp = kprops.get()
        status = start_trace(byref(handle),
                             self._trace_name,
                             kp)
        self._handle = handle
        if status == ERROR_ALREADY_EXISTS:
            # the kernel logger trace session
            # is already running. Restart the trace.
            self.stop_ktrace()
            status = start_trace(byref(handle),
                                 self._trace_name,
                                 kp)
            if status != ERROR_SUCCESS:
                raise FibratusError('Unable to start fibratus')
            self._handle = handle
        elif status == ERROR_ACCESS_DENIED:
            # insufficient privileges
            panic("You don't have administrative privileges. Stopping fibratus...")
        elif status == ERROR_BAD_LENGTH:
            raise FibratusError('Incorrect buffer size for the trace buffer')
        elif status == ERROR_INVALID_PARAMETER:
            raise FibratusError('Invalid trace handle or provider GUID')
        elif status != ERROR_SUCCESS:
            raise FibratusError('Unable to start fibratus')

    def stop_ktrace(self, kprops=None):
        """Stops the current running trace.

        Parameters
        ---------
        kprops: KTraceProps
            an instance of the kernel trace properties
        """
        kprops = kprops or KTraceProps()

        handle = self._handle
        self._handle = TRACEHANDLE()
        control_trace(handle,
                      self._trace_name,
                      kprops.get(),
                      EVENT_TRACE_CONTROL_STOP)



# Copyright 2015 by Nedim Sabic (RabbitStack)
# All Rights Reserved.
# http://rabbitstack.github.io

# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from ctypes import Structure
from ctypes.wintypes import HANDLE, DWORD, BOOL, WCHAR, LONG

import fibratus.apidefs.declarer as declarer
from fibratus.apidefs.cdefs import LPVOID, LPTSTR


FILE_SHARE_READ = 0x00000001
FILE_SHARE_WRITE = 0x00000002
FILE_SHARE_DELETE = 0x00000004

# if the file already exists, replace it with the given file. If it does not, create the given file.
FILE_SUPERSEDE = 0x00000000
# if the file already exists, open it instead of creating a new file.
# If it does not, fail the request and do not create a new file.
FILE_OPEN = 0x00000001
# if the file already exists, fail the request and do not create or open the given file.
# If it does not, create the given file.
FILE_CREATE = 0x00000002
# If the file already exists, open it. If it does not, create the given file.
FILE_OPEN_IF = 0x00000003
# If the file already exists, open it and overwrite it. If it does not, fail the request.
FILE_OVERWRITE = 0x00000004
# If the file already exists, open it and overwrite it. If it does not, create the given file.
FILE_OVERWRITE_IF = 0x00000005

# the file being created or opened is a directory file
FILE_DIRECTORY_FILE = 0x00000001
# open a file with a reparse point and bypass normal reparse point processing for the file
FILE_OPEN_REPARSE_POINT = 0x00200000


class FILE_NAME_INFO(Structure):
    _fields_ = [('file_name_length', DWORD),
                ('filename', WCHAR * 1)]


get_file_info_by_handle = declarer.declare(declarer.KERNEL, 'GetFileInformationByHandleEx',
                                           [HANDLE, DWORD, LPVOID, DWORD],
                                           BOOL)
query_dos_device = declarer.declare(declarer.KERNEL, 'QueryDosDeviceW',
                                    [LPTSTR, LPTSTR, DWORD],
                                    DWORD)

_get_osfhandle = declarer.declare(declarer.C, '_get_osfhandle',
                                  [DWORD],
                                  LONG)

get_file_type = declarer.declare(declarer.KERNEL, 'GetFileType',
                                 [HANDLE],
                                 DWORD)


# Copyright 2016 by Nedim Sabic (RabbitStack)
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import pefile
from fibratus.common import DotD as ddict, underscore_dict_keys


def __decode__(value):
    return value.decode('utf-8')


def __from_idx__(string_table, idx):
    """Lookups the entry in the string table.

    Parameters
    ----------
    string_table: dict
        the string table
    idx: int
        index for the entry
    """
    _, v = string_table[idx]
    return __decode__(v)


class ImageMetaRegistry(object):

    def __init__(self, enabled=True, imports=False, file_info=False):
        """Creates an instace of the image meta registry.

        Arguments
        ---------

        enabled: bool
            determines if image meta information should be added to the registry
        imports: bool
            it instructs the PE module to parse the directory entry import structure
        file_info: bool
            determines if file information meta data should be extracted from the PE
        """
        self.image_metas = {}
        self.imports = imports
        self.file_info = file_info
        self.enabled = enabled
        self.full_loaded = False

    def add_image_meta(self, path):
        """Registers image meta information.

        This method parses the PE (Portable Executable) binary format
        of the the image passed in the `path` parameter.

        It then extracts some basic headers present in the PE, as well
        as sections which form the binary image.

        Parameters
        ----------

        path: str
            the absolute path of the image file
        """
        if not self.enabled:
            return None
        try:
            if (path.endswith('exe') or
                path.endswith('dll') or
                path.endswith('sys')) and \
                 path not in self.image_metas:
                    pe = pefile.PE(path, fast_load=True)
                    file_header = ddict(underscore_dict_keys(pe.FILE_HEADER.dump_dict()))
                    # create image meta instance
                    image_meta = ImageMeta(file_header.machine.value,
                                           file_header.time_date_stamp.value,
                                           file_header.number_of_sections.value)
                    image_meta.sections = [dict(name=__decode__(ddict(se.dump_dict()).Name.Value),
                                                entropy=se.get_entropy(),
                                                md5=se.get_hash_md5(),
                                                sha1=se.get_hash_sha1(),
                                                sha256=se.get_hash_sha256(),
                                                sha512=se.get_hash_sha512())
                                           for se in pe.sections]
                    # parse directory entry imports
                    if self.imports:
                        pe.full_load()
                        self.full_loaded = True
                        for module in self.__directory_entry_import__(pe):
                            dll = __decode__(module.dll)
                            imports = [__decode__(i.name)
                                       for i in module.imports
                                       if not i.import_by_ordinal]
                            image_meta.imports[dll] = imports
                    # parse the string table to extract
                    # the copyright, company, description
                    # and other attributes
                    if self.file_info:
                        if not self.full_loaded:
                            pe.full_load()
                        if self.__pe_has_version_info__(pe):
                            file_info = pe.FileInfo
                            if file_info and len(file_info) > 0:
                                file_info = file_info[0]
                                if self.__fi_has_string_table__(file_info):
                                    string_table = sorted(list(file_info.StringTable[0].entries.items()))
                                    # get file info entries from table index
                                    image_meta.org = __from_idx__(string_table, 0)
                                    image_meta.description = __from_idx__(string_table, 1)
                                    image_meta.version = __from_idx__(string_table, 2)
                                    image_meta.internal_name = __from_idx__(string_table, 3)
                                    image_meta.copyright = __from_idx__(string_table, 4)

                    self.image_metas[path] = image_meta

                    return image_meta
        except Exception:
            # ignore the exception for now
            # but consider logging it to file
            # in case it can provide hints for
            # troubleshooting purposes
            pass

    def get_image_meta(self, path):
        return self.image_metas[path] if path in self.image_metas else None

    def remove_image_meta(self, path):
        return self.image_metas.pop(path, None)

    def __pe_has_version_info__(self, pe):
        return hasattr(pe, 'VS_VERSIONINFO')

    def __fi_has_string_table__(self, file_info):
        return len(file_info.StringTable) > 0 and hasattr(file_info, 'StringTable')

    def __directory_entry_import__(self, pe):
        return getattr(pe, 'DIRECTORY_ENTRY_IMPORT') if hasattr(pe, 'DIRECTORY_ENTRY_IMPORT') else []


class ImageMeta(object):
    """Container for a plethora of metadata extracted from the PE headers.

    Attributes
    ----------
    arch: str
        identifies the target architecture for which this image is compiled
    timestamp: str
        the date and time the image was created by the linker
    num_sections: int
        indicates the size of the section table
    sections: list
        information for every section found in the image
    """
    def __init__(self, arch, timestamp, num_sections):
        self._arch = 'x86-64' if arch == 34404 else 'x86'
        self._timestamp = timestamp
        self._num_sections = num_sections
        self._sections = []
        self._org = None
        self._description = None
        self._version = None
        self._internal_name = None
        self._copyright = None

        self._imports = {}

    @property
    def arch(self):
        return self._arch

    @property
    def timestamp(self):
        return self._timestamp

    @property
    def num_sections(self):
        return self._num_sections

    @property
    def org(self):
        return self._org

    @org.setter
    def org(self, org):
        self._org = org

    @property
    def description(self):
        return self._description

    @description.setter
    def description(self, description):
        self._description = description

    @property
    def version(self):
        return self._version

    @version.setter
    def version(self, version):
        self._version = version

    @property
    def internal_name(self):
        return self._internal_name

    @internal_name.setter
    def internal_name(self, internal_name):
        self._internal_name = internal_name

    @property
    def copyright(self):
        return self._copyright

    @copyright.setter
    def copyright(self, copyright):
        self._copyright = copyright

    @property
    def sections(self):
        return self._sections

    @sections.setter
    def sections(self, sections):
        self._sections = sections

    @property
    def imports(self):
        return self._imports



# Copyright 2016 by Nedim Sabic (RabbitStack)
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import re
import os
import shutil
from os.path import expanduser
from setuptools import setup, find_packages
from setuptools.extension import Extension
from Cython.Distutils import build_ext
from pip.req import parse_requirements

from fibratus.version import VERSION


def copy_artifacts():
    home_dir = expanduser('~')
    here = os.path.abspath(os.path.dirname(__file__))
    fibratus_dir = os.path.join(home_dir, '.fibratus')
    if not os.path.exists(fibratus_dir):
        os.mkdir(fibratus_dir)
        shutil.copy(os.path.join(here, 'fibratus.yml'), fibratus_dir)
        shutil.copy(os.path.join(here, 'schema.yml'), fibratus_dir)
        shutil.copytree(os.path.join(here, 'filaments'), os.path.join(fibratus_dir, 'filaments'))

kstreamc_ext = Extension('kstreamc',
                         ['kstream/kstreamc.pyx'],
                         libraries=["tdh", "advapi32", "ole32", "ws2_32"],
                         language='c++')

install_reqs = parse_requirements('requirements.txt', session=False)
reqs = [str(ir.req) for ir in install_reqs if not re.match('pytest|codecov', str(ir.req))]

copy_artifacts()

setup(
    name="fibratus",
    version=VERSION,
    author="Nedim Sabic (RabbitStack)",
    author_email="bhnedo@hotmail.com",
    description="Tool for exploration and tracing of the Windows kernel",
    long_description="Fibratus is a tool which is able to capture the most of the Windows kernel activity - "
                     "process/thread creation and termination, file system I/O, registry, network activity, "
                     "DLL loading/unloading and much more. Fibratus has a very simple CLI which encapsulates "
                     "the machinery to start the kernel event stream collector, set kernel event filters or "
                     "run the lightweight Python modules called filaments. You can use filaments to extend "
                     "Fibratus with your own arsenal of tools.",
    license="Apache",
    keywords="windows kernel, tracing, system exploration, syscalls",
    platforms=["Windows"],
    url="https://github.com/rabbitstack/fibratus",
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Topic :: System',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4'
    ],
    ext_modules=[kstreamc_ext],
    cmdclass={"build_ext": build_ext},
    packages=find_packages(),
    install_requires=reqs,
    entry_points={
        'console_scripts': [
            'fibratus=fibratus.cli:main',
        ],
    }
)
